# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import random
import numpy as np
import nevergrad as ng
import nevergrad.common.typing as tp
from nevergrad.parametrization import parameter as p
from . import nsga2

# pylint: disable=reimported,redefined-outer-name,unused-variable,unsubscriptable-object, unused-argument
# pylint: disable=import-outside-toplevel


def test_crowding_distance() -> None:
    params = ng.p.Tuple(ng.p.Scalar(lower=0, upper=2), ng.p.Scalar(lower=0, upper=2))

    candidates: tp.List[p.Parameter] = []
    v = sorted([random.uniform(0.01, 0.99) for i in range(4)])
    loss_values = [[0.0, 5.0], [v[0], v[3]], [v[1], v[2]], [1.0, 0.0]]
    for i, v in enumerate(loss_values):
        candidates.append(params.spawn_child().set_standardized_data(v))
        candidates[i]._losses = np.array(v)
    crowding_distance = nsga2.CrowdingDistance()
    crowding_distance.compute_distance(candidates)

    # For objective 1
    cdist_1 = (loss_values[2][0] - loss_values[0][0]) / abs(loss_values[0][0] - loss_values[-1][0])
    cdist_2 = (loss_values[3][0] - loss_values[1][0]) / abs(loss_values[0][0] - loss_values[-1][0])

    # For objective 2
    cdist_1 += (loss_values[0][1] - loss_values[2][1]) / abs(loss_values[0][1] - loss_values[-1][1])
    cdist_2 += (loss_values[1][1] - loss_values[3][1]) / abs(loss_values[0][1] - loss_values[-1][1])

    assert candidates[0]._meta["crowding_distance"] == float("inf")
    np.testing.assert_almost_equal(candidates[1]._meta["crowding_distance"], cdist_1, decimal=3)
    np.testing.assert_almost_equal(candidates[2]._meta["crowding_distance"], cdist_2, decimal=3)
    assert candidates[3]._meta["crowding_distance"] == float("inf")


def test_fast_non_dominated_ranking() -> None:
    params = ng.p.Tuple(ng.p.Scalar(lower=0, upper=2), ng.p.Scalar(lower=0, upper=2))

    loss_values = [[[0.0, 2.0], [1.0, 1.0]], [[0.0, 4.0], [1.0, 3.0], [3.0, 1.0]], [[2.0, 3.0], [4.0, 2.0]]]

    candidates: tp.List[p.Parameter] = []
    expected_frontiers = []
    for vals in loss_values:
        expected_frontier = []
        for v in vals:
            candidate = params.spawn_child().set_standardized_data(v)
            candidate._losses = np.array(v)
            candidates.append(candidate)
            expected_frontier.append(candidate)
        expected_frontiers.append(expected_frontier)

    ranking_method = nsga2.FastNonDominatedRanking()
    frontiers = ranking_method.compute_ranking(candidates)
    assert set(frontiers[0]) == set(expected_frontiers[0])
    assert set(frontiers[1]) == set(expected_frontiers[1])
    assert set(frontiers[2]) == set(expected_frontiers[2])


def get_nsga2_test_case_data():
    params = ng.p.Tuple(ng.p.Scalar(lower=0, upper=2), ng.p.Scalar(lower=0, upper=2))

    loss_values = [[[0.0, 2.0], [1.0, 1.0]], [[0.0, 4.0], [1.0, 3.0], [3.0, 1.0]], [[2.0, 3.0], [4.0, 2.0]]]

    candidates: tp.List[p.Parameter] = []
    expected_frontiers = []
    for vals in loss_values:
        expected_frontier = []
        for v in vals:
            candidate = params.spawn_child().set_standardized_data(v)
            candidate._losses = np.array(v)
            candidates.append(candidate)
            expected_frontier.append(candidate)
        expected_frontiers.append(expected_frontier)
    return candidates, expected_frontiers


def test_nsga2_ranking() -> None:
    candidates, expected_frontiers = get_nsga2_test_case_data()
    rank_result = nsga2.rank(candidates, len(candidates))

    assert len(rank_result) == len(candidates)
    for i, frontier in enumerate(expected_frontiers):
        for c in frontier:
            assert rank_result[c.uid][0] == i


def test_nsga2_ranking_2() -> None:
    candidates, expected_frontiers = get_nsga2_test_case_data()
    n_selected = len(expected_frontiers[0]) + len(expected_frontiers[1]) - 1
    rank_result = nsga2.rank(candidates, n_selected)

    assert len(rank_result) == n_selected
    # Check the first frontier
    max_rank_frontier1 = 0
    for c in expected_frontiers[0]:
        assert rank_result[c.uid][2] == float("inf")
        max_rank_frontier1 = max(max_rank_frontier1, rank_result[c.uid][0])

    # Check the second frontier
    n_cand_in_frontier2 = 0
    for c in expected_frontiers[1]:
        if c.uid in rank_result:
            n_cand_in_frontier2 += 1
            assert rank_result[c.uid][0] > max_rank_frontier1
    assert n_cand_in_frontier2 == len(expected_frontiers[1]) - 1


def test_nsga2_ranking_3() -> None:
    candidates, expected_frontiers = get_nsga2_test_case_data()
    rank_result = nsga2.rank(candidates, None)

    assert len(rank_result) == len(candidates)
    for i, frontier in enumerate(expected_frontiers):
        expect_n_non_inf = max(0, len(frontier) - 2)
        n_non_inf = 0
        for c in frontier:
            assert rank_result[c.uid][1] == i
            if rank_result[c.uid][2] != float("inf"):
                n_non_inf += 1
        assert n_non_inf == expect_n_non_inf


def test_nsga2_ranking_4():
    params = ng.p.Tuple(ng.p.Scalar(lower=0, upper=2))
    loss_values = [0.0, 1.0, -10.0, 1.0, 3.0, 1.0]
    candidates: tp.List[p.Parameter] = []
    for v in loss_values:
        candidate = params.spawn_child().set_standardized_data(v)
        candidate.loss = np.array(v)
        candidates.append(candidate)

    n_selected = 3
    rank_result = nsga2.rank(candidates, n_selected)
    candidates.sort(key=lambda x: rank_result[x.uid][0] if x.uid in rank_result else float("inf"))
    loss_from_rank = [r.loss for r in candidates[:n_selected]]
    loss_from_sorted = [np.array(v) for v in sorted(loss_values)[:n_selected]]
    assert loss_from_rank == loss_from_sorted



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import re
import string
import hashlib
import warnings
import argparse
import itertools
from collections import defaultdict
from pathlib import Path
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.legend import Legend
from matplotlib import cm
from mpl_toolkits.axes_grid1 import make_axes_locatable
import nevergrad.common.typing as tp
from . import utils
from .exporttable import export_table

# pylint: disable=too-many-locals


_DPI = 250
pure_algorithms = []

# %% Basic tools


def compactize(name: str) -> str:
    if len(name) < 70:
        return name
    hashcode = hashlib.md5(bytes(name, "utf8")).hexdigest()
    name = re.sub(r"\([^()]*\)", "", name)
    mid = 35
    name = name[:mid] + hashcode + name[-mid:]
    return name


def _make_style_generator() -> tp.Iterator[str]:
    lines = itertools.cycle(["-", "--", ":", "-."])  # 4
    markers = itertools.cycle("ov^<>8sp*hHDd")  # 13
    colors = itertools.cycle("bgrcmyk")  # 7
    return (l + m + c for l, m, c in zip(lines, markers, colors))


class NameStyle(tp.Dict[str, tp.Any]):
    """Provides a style for each name, and keeps to it"""

    def __init__(self) -> None:
        super().__init__()
        self._gen = _make_style_generator()

    def __getitem__(self, name: str) -> tp.Any:
        if name not in self:
            super().__setitem__(name, next(self._gen))
        return super().__getitem__(name)


def _make_winners_df(df: pd.DataFrame, all_optimizers: tp.List[str]) -> utils.Selector:
    """Finds mean loss over all runs for each of the optimizers, and creates a matrix
    winner_ij = 1 if opt_i is better (lower loss) then opt_j (and .5 for ties)
    """
    if not isinstance(df, utils.Selector):
        df = utils.Selector(df)
    all_optim_set = set(all_optimizers)
    assert all(x in all_optim_set for x in df.unique("optimizer_name"))
    assert all(x in df.columns for x in ["optimizer_name", "loss"])
    winners = utils.Selector(index=all_optimizers, columns=all_optimizers, data=0.0)
    grouped = df.loc[:, ["optimizer_name", "loss"]].groupby(["optimizer_name"]).mean()
    df_optimizers = list(grouped.index)
    values = np.array(grouped)
    diffs = values - values.T
    # loss_ij = 1 means opt_i beats opt_j once (beating means getting a lower loss/regret)
    winners.loc[df_optimizers, df_optimizers] = (diffs < 0) + 0.5 * (diffs == 0)
    return winners


def aggregate_winners(
    df: utils.Selector, categories: tp.List[str], all_optimizers: tp.List[str]
) -> tp.Tuple[utils.Selector, int]:
    """Computes the sum of winning rates on all cases corresponding to the categories

    Returns
    -------
    Selector
        the aggregate
    int
        the total number of cases
    """
    if not categories:
        return _make_winners_df(df, all_optimizers), 1
    subcases = df.unique(categories[0])
    if len(subcases) == 1:
        return aggregate_winners(df, categories[1:], all_optimizers)
    iterdf, iternum = zip(
        *(
            aggregate_winners(
                df.loc[
                    df.loc[:, categories[0]] == val
                    if categories[0] != "budget"
                    else df.loc[:, categories[0]] <= val
                ],
                categories[1:],
                all_optimizers,
            )
            for val in subcases
        )
    )
    return sum(iterdf), sum(iternum)  # type: ignore


def _make_sorted_winrates_df(victories: pd.DataFrame) -> pd.DataFrame:
    """Converts a dataframe counting number of victories into a sorted
    winrate dataframe. The algorithm which performs better than all other
    algorithms comes first. When you do not play in a category, you are
    considered as having lost all comparisons in that category.
    """
    assert all(x == y for x, y in zip(victories.index, victories.columns))
    winrates = victories / (victories + victories.T).max(axis=1)
    # mean_win = winrates.quantile(.05, axis=1).sort_values(ascending=False)
    mean_win = winrates.mean(axis=1).sort_values(ascending=False)
    return winrates.loc[mean_win.index, mean_win.index]


# %% plotting functions


def remove_errors(df: pd.DataFrame) -> utils.Selector:
    df = utils.Selector(df)
    if "error" not in df.columns:  # backward compatibility
        return df  # type: ignore
    # errors with no recommendation
    nandf = df.select(loss=np.isnan)
    for row in nandf.itertuples():
        msg = f'Removing "{row.optimizer_name}"'
        msg += f" with dimension {row.dimension}" if hasattr(row, "dimension") else ""
        msg += f': got error "{row.error}"' if isinstance(row.error, str) else "recommended a nan"
        warnings.warn(msg)
    # error with recorded recommendation
    handlederrordf = df.select(error=lambda x: isinstance(x, str) and x, loss=lambda x: not np.isnan(x))
    for row in handlederrordf.itertuples():
        warnings.warn(
            f'Keeping non-optimal recommendation of "{row.optimizer_name}" '
            f'with dimension {row.dimension if hasattr(row, "dimension") else "UNKNOWN"} which raised "{row.error}".'
        )
    err_inds = set(nandf.index)
    output = df.loc[[i for i in df.index if i not in err_inds], [c for c in df.columns if c != "error"]]
    # cast nans in loss to infinity
    df.loc[np.isnan(df.loss), "loss"] = float("inf")
    #
    assert (
        not output.loc[:, "loss"].isnull().values.any()
    ), "Some nan values remain while there should not be any!"
    output = utils.Selector(output.reset_index(drop=True))
    return output  # type: ignore


class PatternAggregate:
    def __init__(self, pattern: str) -> None:
        self._pattern = pattern

    def __call__(self, df: pd.Series) -> str:
        return self._pattern.format(**df.to_dict())


_PARAM_MERGE_PATTERN = "{optimizer_name},{parametrization}"


def merge_optimizer_name_pattern(
    df: utils.Selector, pattern: str, merge_parametrization: bool = False, remove_suffix: bool = False
) -> utils.Selector:
    """Merge the optimizer name with other descriptors based on a pattern
    Nothing happens if merge_parametrization is false and pattern is empty string
    """
    if merge_parametrization:
        if pattern:
            raise ValueError(
                "Cannot specify both merge-pattern and merge-parametrization "
                "(merge-parametrization is equivalent to merge-pattern='{optimizer_name},{parametrization}')"
            )
        pattern = _PARAM_MERGE_PATTERN
    if not pattern:
        return df
    df = df.copy()
    okey = "optimizer_name"
    elements = [tup[1] for tup in string.Formatter().parse(pattern) if tup[1] is not None]
    assert okey in elements, (
        f"Missing optimizer key {okey!r} in merge pattern.\nEg: "
        + 'pattern="{optimizer_name}_{parametrization}"'
    )
    others = [x for x in elements if x != okey]
    aggregate = PatternAggregate(pattern)
    sub = df.loc[:, elements].fillna("")
    if len(sub.unique(others)) > 1:
        for optim in sub.unique(okey):
            inds = sub.loc[:, okey] == optim
            if len(sub.loc[inds, :].unique(others)) > 1:
                df.loc[inds, okey] = sub.loc[inds, elements].agg(aggregate, axis=1)
    if remove_suffix:
        df["optimizer_name"] = df["optimizer_name"].replace(r"[0-9\.\-]*$", "", regex=True)
    return df.drop(columns=others)  # type: ignore


def normalized_losses(df: pd.DataFrame, descriptors: tp.List[str]) -> utils.Selector:
    df = utils.Selector(df.copy())
    cases = df.unique(descriptors)
    if not cases:
        cases = [()]
    # Average normalized plot with everything.
    for case in cases:
        subdf = df.select_and_drop(**dict(zip(descriptors, case)))
        losses = np.array(subdf.loc[:, "loss"])
        m = min(losses)
        M = max(losses[losses < float("inf")])
        df.loc[subdf.index, "loss"] = (df.loc[subdf.index, "loss"] - m) / (M - m) if M != m else 1
    return df  # type: ignore


# pylint: disable=too-many-statements,too-many-branches
def create_plots(
    df: pd.DataFrame,
    output_folder: tp.PathLike,
    max_combsize: int = 1,
    xpaxis: str = "budget",
    competencemaps: bool = False,
    nomanyxp: bool = False,
) -> None:
    """Saves all representing plots to the provided folder

    Parameters
    ----------
    df: pd.DataFrame
        the experiment data
    output_folder: PathLike
        path of the folder where the plots should be saved
    max_combsize: int
        maximum number of parameters to fix (combinations) when creating experiment plots
    xpaxis: str
        x-axis for xp plots (either budget or pseudotime)
    """
    assert xpaxis in ["budget", "pseudotime"]
    df = remove_errors(df)
    df.loc[:, "loss"] = pd.to_numeric(df.loc[:, "loss"])
    df = df.loc[:, [x for x in df.columns if not x.startswith("info/")]]
    # Normalization of types.
    for col in df.columns:
        if "max_irr" in col:
            df[col] = df[col].round(decimals=4)
        if col in (
            "budget",
            "num_workers",
            "dimension",
            "useful_dimensions",
            "num_blocks",
            "block_dimension",
            "num_objectives",
        ):
            try:
                df[col] = df[col].astype(float).astype(int)
            except Exception as e1:
                try:
                    for i in range(len(df[col])):
                        float(df[col][i])
                except Exception as e2:
                    assert False, f"Fails at row {i+2}, Exceptions: {e1}, {e2}"
        elif col != "loss":
            df[col] = df[col].astype(str)
            df[col] = df[col].replace(r"\.[0]*$", "", regex=True)
            try:
                df.loc[:, col] = pd.to_numeric(df.loc[:, col])
            except:
                pass
    if "num_objectives" in df.columns:
        df = df[df.num_objectives != 0]  # the optimization did not even start
    # If we have a descriptor "instrum_str",
    # we assume that it describes the instrumentation as a string,
    # that we should include the various instrumentations as distinct curves in the same plot.
    # So we concat it at the end of the optimizer name, and we remove "parametrization"
    # from the descriptor.
    if "instrum_str" in set(df.columns):
        df.loc[:, "optimizer_name"] = df.loc[:, "optimizer_name"] + df.loc[:, "instrum_str"]
        df = df.drop(columns="instrum_str")
        df = df.drop(columns="dimension")
        if "parametrization" in set(df.columns):
            df = df.drop(columns="parametrization")
        if "instrumentation" in set(df.columns):
            df = df.drop(columns="instrumentation")
    df = utils.Selector(df.fillna("N-A"))  # remove NaN in non score values
    assert not any("Unnamed: " in x for x in df.columns), f"Remove the unnamed index column:  {df.columns}"
    assert "error " not in df.columns, f"Remove error rows before plotting"
    required = {"optimizer_name", "budget", "loss", "elapsed_time", "elapsed_budget"}
    missing = required - set(df.columns)
    assert not missing, f"Missing fields: {missing}"
    output_folder = Path(output_folder)
    os.makedirs(output_folder, exist_ok=True)
    # check which descriptors do vary
    descriptors = sorted(
        set(df.columns) - (required | {"instrum_str", "seed", "pseudotime"})
    )  # all other columns are descriptors
    to_drop = [x for x in descriptors if len(df.unique(x)) == 1]
    df = utils.Selector(df.loc[:, [x for x in df.columns if x not in to_drop]])
    # now those should be actual interesting descriptors
    all_descriptors = sorted(set(df.columns) - (required | {"instrum_str", "seed", "pseudotime"}))
    print(f"Descriptors: {all_descriptors}")
    print("# Fight plots")
    #
    # fight plot
    # choice of the combination variables to fix
    fight_descriptors = all_descriptors + ["budget"]  # budget can be used as a descriptor for fight plots
    combinable = [x for x in fight_descriptors if len(df.unique(x)) > 1]  # should be all now
    # We remove descriptors which have only one value for each budget.
    descriptors = []
    for d in all_descriptors:
        acceptable = False
        for b in df.budget.unique():
            if len(df.loc[df["budget"] == b][d].unique()) > 1:
                acceptable = True
                break
        if acceptable:
            descriptors += [d]
    num_rows = 6

    # For the competence map case we must consider pairs of attributes, hence maxcomb_size >= 2.
    # A competence map shows for each value of each of two attributes which algorithm was best.
    if competencemaps:
        max_combsize = max(max_combsize, 2)
    for fixed in list(
        itertools.chain.from_iterable(
            itertools.combinations(combinable, order) for order in range(max_combsize + 1)
        )
    ):
        orders = [len(c) for c in df.unique(fixed)]
        if orders:
            assert min(orders) == max(orders)
            order = min(orders)
        else:
            order = 0
        best_algo: tp.List[tp.List[str]] = []
        if competencemaps and order == 2:  # With order 2 we can create a competence map.
            print("\n#trying to competence-map")
            if all(
                [len(c) > 1 for c in df.unique(fixed)]
            ):  # Let us try if data are adapted to competence maps.
                # This is not always the case, as some attribute1/value1 + attribute2/value2 might be empty
                # (typically when attribute1 and attribute2 are correlated).
                try:
                    xindices = sorted(set(c[0] for c in df.unique(fixed)))
                except TypeError:
                    xindices = list(set(c[0] for c in df.unique(fixed)))
                try:
                    yindices = sorted(set(c[1] for c in df.unique(fixed)))
                except TypeError:
                    yindices = list(set(c[1] for c in df.unique(fixed)))
                for _ in range(len(xindices)):
                    best_algo += [[]]
                for i in range(len(xindices)):
                    for _ in range(len(yindices)):
                        best_algo[i] += ["none"]

        # Let us loop over all combinations of variables.
        for case in df.unique(fixed) if fixed else [()]:
            print("\n# new case #", fixed, case)
            casedf = df.select(**dict(zip(fixed, case)))
            data_df = FightPlotter.winrates_from_selection(
                casedf, fight_descriptors, num_rows=num_rows, num_cols=350
            )
            fplotter = FightPlotter(data_df)
            # Competence maps: we find out the best algorithm for each attribute1=valuei/attribute2=valuej.
            if order == 2 and competencemaps and best_algo:
                print("\n#storing data for competence-map")
                best_algo[xindices.index(case[0])][yindices.index(case[1])] = fplotter.winrates.index[0]
            # save
            name = "fight_" + ",".join("{}{}".format(x, y) for x, y in zip(fixed, case)) + ".png"
            name = "fight_all.png" if name == "fight_.png" else name
            name = compactize(name)
            fullname = name

            if len(name) > 240:
                hashcode = hashlib.md5(bytes(name, "utf8")).hexdigest()
                name = re.sub(r"\([^()]*\)", "", name)
                mid = 120
                name = name[:mid] + hashcode + name[-mid:]
            fplotter.save(str(output_folder / name), dpi=_DPI)
            # Second version, restricted to cases with all data available.
            data_df = FightPlotter.winrates_from_selection(
                casedf, fight_descriptors, num_rows=num_rows, complete_runs_only=True
            )
            fplotter = FightPlotter(data_df)
            if name == "fight_all.png":
                with open(str(output_folder / name) + ".cp.txt", "w") as f:
                    f.write(fullname)
                    f.write("ranking:\n")
                    global pure_algorithms
                    pure_algorithms = list(data_df.columns[:])
                    for i, algo in enumerate(data_df.columns[:158]):
                        f.write(f"  algo {i}: {algo}\n")
            if name == "fight_all.png":
                fplotter.save(str(output_folder / "fight_all_pure.png"), dpi=_DPI)
            else:
                fplotter.save(str(output_folder / name) + "_pure.png", dpi=_DPI)
                print(f"# {len(data_df.columns[:])}  {data_df.columns[:]}")
            if order == 2 and competencemaps and best_algo:  # With order 2 we can create a competence map.
                print("\n# Competence map")
                name = "competencemap_" + ",".join("{}".format(x) for x in fixed) + ".tex"
                export_table(str(output_folder / name), xindices, yindices, best_algo)
                print("Competence map data:", fixed, case, best_algo)

    plt.close("all")
    # xp plots: for each experimental setup, we plot curves with budget in x-axis.
    # plot mean loss / budget for each optimizer for 1 context
    print("# Xp plots")
    name_style = NameStyle()  # keep the same style for each algorithm
    cases = df.unique(descriptors)
    if not cases:
        cases = [()]
    # Average normalized plot with everything.
    out_filepath = output_folder / "xpresults_all.png"
    data = XpPlotter.make_data(df, normalized_loss=True)
    xpplotter = XpPlotter(
        data, title=os.path.basename(output_folder), name_style=name_style, xaxis=xpaxis, pure_only=True
    )
    xpplotter.save(out_filepath)
    # Now one xp plot per case.
    for case in cases:
        if nomanyxp:
            continue
        subdf = df.select_and_drop(**dict(zip(descriptors, case)))
        description = ",".join("{}:{}".format(x, y) for x, y in zip(descriptors, case))
        full_description = description
        description = compactize(description)
        if len(description) > 280:
            hash_ = hashlib.md5(bytes(description, "utf8")).hexdigest()
            description = description[:140] + hash_ + description[-140:]
        out_filepath = output_folder / "xpresults{}{}.png".format(
            "_" if description else "", description.replace(":", "")
        )
        txt_out_filepath = output_folder / "xpresults{}{}.leaderboard.txt".format(
            "_" if description else "", description.replace(":", "")
        )
        data = XpPlotter.make_data(subdf)
        try:
            xpplotter = XpPlotter(data, title=description, name_style=name_style, xaxis=xpaxis)
        except Exception as e:  # pylint: disable=broad-except
            warnings.warn(f"Bypassing error in xpplotter:\n{e}", RuntimeWarning)
        else:
            xpplotter.save(out_filepath)
            xpplotter.save_txt(txt_out_filepath, data, full_description)
    plt.close("all")


def gp_sota() -> tp.Dict[str, tp.Tuple[float, float]]:
    gp = {}
    gp["CartPole-v1"] = (-500.0, 100000.0)
    gp["Acrobot-v1"] = (83.17, 200000.0)
    gp["MountainCarContinuous-v0"] = (-99.31, 900000.0)
    gp["Pendulum-v0"] = (154.36, 1100000.0)
    gp["InvertedPendulumSwingupBulletEnv-v0"] = (-893.35, 400000.0)
    gp["BipedalWalker-v3"] = (-268.85, 1100000.0)
    gp["BipedalWalkerHardcore-v3"] = (-9.25, 1100000.0)
    gp["HopperBulletEnv-v0"] = (-999.19, 1000000.0)
    gp["InvertedDoublePendulumBulletEnv-v0"] = (-9092.17, 300000.0)
    gp["LunarLanderContinuous-v2"] = (-287.58, 1000000.0)
    return gp


class LegendInfo(tp.NamedTuple):
    """Handle for information used to create a legend."""

    x: float
    y: float
    line: tp.Any
    text: str


class XpPlotter:
    """Creates a xp result plot out of the given dataframe: regret with respect to budget for
    each optimizer after averaging on all experiments (it is good practice to use a df
    which is filtered out for one set of input parameters)

    Parameters
    ----------
    optim_vals: dict
        output of the make_data static method, containing all information necessary for plotting
    title: str
        title of the plot
    name_style: dict
        a dict or dict-like object providing a line style for each optimizer name.
        (can be helpful for consistency across plots)
    """

    def __init__(
        self,
        optim_vals: tp.Dict[str, tp.Dict[str, np.ndarray]],
        title: str,
        name_style: tp.Optional[tp.Dict[str, tp.Any]] = None,
        xaxis: str = "budget",
        pure_only: bool = False,
    ) -> None:
        if name_style is None:
            name_style = NameStyle()
        upperbound = max(
            np.max(vals["loss"]) for vals in optim_vals.values() if np.max(vals["loss"]) < np.inf
        )
        for optim, vals in optim_vals.items():
            if optim.lower() in ["stupid", "idiot"] or optim in ["Zero", "StupidRandom"]:
                upperbound = min(upperbound, np.max(vals["loss"]))
        # plot from best to worst
        lowerbound = np.inf
        sorted_optimizers = sorted(optim_vals, key=lambda x: optim_vals[x]["loss"][-1], reverse=True)
        if pure_only:
            assert len(pure_algorithms) > 0
            sorted_optimizers = [
                o for o in sorted_optimizers if o + " " in [p[: (len(o) + 1)] for p in pure_algorithms]
            ]
            with open("rnk__" + str(title) + ".cp.txt", "w") as f:
                f.write(compactize(title))
                f.write("ranking:\n")
                for i, algo in reversed(list(enumerate(sorted_optimizers))):
                    f.write(f"  algo {i}: {algo} (x)\n")
            # print(sorted_optimizers, " merged with ", pure_algorithms)
            # print("Leads to ", sorted_optimizers)
        self._fig = plt.figure()
        self._ax = self._fig.add_subplot(111)
        # use log plot? yes, if no negative value
        logplot = not any(
            x <= 0 or x > 10**8 for ov in optim_vals.values() for x in ov["loss"]
        )  # if x < np.inf)
        if logplot:
            self._ax.set_yscale("log")
            for ov in optim_vals.values():
                if ov["loss"].size:
                    ov["loss"] = np.maximum(1e-30, ov["loss"])
        # other setups
        self._ax.autoscale(enable=False)
        self._ax.set_xscale("log")
        self._ax.set_xlabel(xaxis)
        self._ax.set_ylabel("loss")
        self._ax.grid(True, which="both")
        self._overlays: tp.List[tp.Any] = []
        legend_infos: tp.List[LegendInfo] = []
        title_addendum = ""
        for optim_name in (
            sorted_optimizers[:1] + sorted_optimizers[-12:]
            if len(sorted_optimizers) > 13
            else sorted_optimizers
        ):
            vals = optim_vals[optim_name]
            indices = np.where(vals["num_eval"] > 0)
            lowerbound = min(lowerbound, np.min(vals["loss"]))
            # We here add some state of the art results.
            # This adds a cross on figures, x-axis = budget and y-axis = loss.
            for sota_name, sota in [("GP", gp_sota())]:
                for k in sota.keys():
                    if k in title:
                        th = sota[k][0]  # loss of proposed solution.
                        cost = sota[k][1]  # Computational cost for the proposed result.
                        title_addendum = f"({sota_name}:{th})"
                        lowerbound = min(lowerbound, th, 0.9 * th, 1.1 * th)
                        plt.plot(  # Horizontal line at the obtained GP cost.
                            vals[xaxis][indices],
                            th + 0 * vals["loss"][indices],
                            name_style[optim_name],
                            label="gp",
                        )
                        plt.plot(  # Vertical line, showing the budget of the GP solution.
                            [cost] * 3,
                            [
                                min(vals["loss"][indices]),
                                sum(vals["loss"][indices]) / len(indices),
                                max(vals["loss"][indices]),
                            ],
                            name_style[optim_name],
                            label="gp",
                        )
            line = plt.plot(vals[xaxis], vals["loss"], name_style[optim_name], label=optim_name)
            # confidence lines
            for conf in self._get_confidence_arrays(vals, log=logplot):
                plt.plot(vals[xaxis], conf, name_style[optim_name], label=optim_name, alpha=0.1)
            text = "{} ({:.3g} <{:.3g}>)".format(
                optim_name,
                vals["loss"][-1],
                vals["loss"][-2] if len(vals["loss"]) > 2 else float("nan"),
            )
            if vals[xaxis].size:
                legend_infos.append(LegendInfo(vals[xaxis][-1], vals["loss"][-1], line, text))
        if not (np.isnan(upperbound) or np.isinf(upperbound)):
            upperbound_up = upperbound
            if not (np.isnan(lowerbound) or np.isinf(lowerbound)):
                self._ax.set_ylim(bottom=lowerbound)
                upperbound_up += 0.02 * (upperbound - lowerbound)
                if logplot:
                    upperbound_up = 10 ** (
                        np.log10(upperbound) + 0.02 * (np.log10(upperbound) - np.log10(lowerbound))
                    )
            self._ax.set_ylim(top=upperbound_up)
        all_x = [v for vals in optim_vals.values() for v in vals[xaxis]]
        self._ax.set_xlim([min(all_x), max(all_x)])
        self.add_legends(legend_infos)
        # global info
        if "tmp" not in title:
            self._ax.set_title(split_long_title(title + title_addendum))
        self._ax.tick_params(axis="both", which="both")
        # self._fig.tight_layout()

    @staticmethod
    def _get_confidence_arrays(
        vals: tp.Dict[str, np.ndarray], log: bool = False
    ) -> tp.Tuple[np.ndarray, np.ndarray]:
        loss = vals["loss"]
        conf = vals["loss_std"] / np.sqrt(vals["num_eval"] - 1)
        if not log:
            return loss - conf, loss + conf
        lloss = np.log10(loss)
        lstd = 0.434 * conf / loss
        return tuple(10 ** (lloss + x) for x in [-lstd, lstd])  # type: ignore

    def add_legends(self, legend_infos: tp.List[LegendInfo]) -> None:
        """Adds the legends"""
        ax = self._ax
        trans = ax.transScale + ax.transLimits
        fontsize = 10.0
        display_y = (ax.transAxes.transform((1, 1)) - ax.transAxes.transform((0, 0)))[1]  # height in points
        shift = (2.0 + fontsize) / display_y
        legend_infos = legend_infos[::-1]  # revert order for use in compute_best_placements
        values = [float(np.clip(trans.transform((0, i.y))[1], -0.01, 1.01)) for i in legend_infos]
        placements = compute_best_placements(values, min_diff=shift)
        for placement, info in zip(placements, legend_infos):
            self._overlays.append(
                Legend(
                    ax,
                    info.line,
                    [info.text],
                    loc="center left",
                    bbox_to_anchor=(1, placement),
                    frameon=False,
                    fontsize=fontsize,
                )
            )
            ax.add_artist(self._overlays[-1])

    @staticmethod
    def make_data(df: pd.DataFrame, normalized_loss: bool = False) -> tp.Dict[str, tp.Dict[str, np.ndarray]]:
        """Process raw xp data and process it to extract relevant information for xp plots:
        regret with respect to budget for each optimizer after averaging on all experiments (it is good practice to use a df
        which is filtered out for one set of input parameters)

        Parameters
        ----------
        df: pd.DataFrame
            run data
        normalized_loss: bool
            whether we should normalize each data (for each budget and run) between 0 and 1. Convenient when we consider
            averages over several distinct functions that can have very different ranges - then we return data which are rescaled to [0,1].
            Warning: then even if algorithms converge (i.e. tend to minimize), the value can increase, because the normalization
            is done separately for each budget.
        """
        if normalized_loss:
            descriptors = sorted(
                set(df.columns)
                - {
                    "pseudotime",
                    "time",
                    "budget",
                    "elapsed_time",
                    "elapsed_budget",
                    "loss",
                    "optimizer_name",
                    "seed",
                }
            )
            df = normalized_losses(df, descriptors=descriptors)
        df = utils.Selector(
            df.loc[
                :,
                ["optimizer_name", "budget", "loss"] + (["pseudotime"] if "pseudotime" in df.columns else []),
            ]
        )
        groupeddf = df.groupby(["optimizer_name", "budget"])
        means = groupeddf.mean()
        stds = groupeddf.std()
        optim_vals: tp.Dict[str, tp.Dict[str, np.ndarray]] = {}
        # extract name and coordinates
        for optim in df.unique("optimizer_name"):
            optim_vals[optim] = {}
            optim_vals[optim]["budget"] = np.array(means.loc[optim, :].index)
            optim_vals[optim]["loss"] = np.array(means.loc[optim, "loss"])
            optim_vals[optim]["loss_std"] = np.array(stds.loc[optim, "loss"])
            num_eval = np.array(groupeddf.count().loc[optim, "loss"])
            optim_vals[optim]["num_eval"] = num_eval
            if "pseudotime" in means.columns:
                optim_vals[optim]["pseudotime"] = np.array(means.loc[optim, "pseudotime"])
        return optim_vals

    @staticmethod
    def save_txt(
        output_filepath: tp.PathLike, optim_vals: tp.Dict[str, tp.Dict[str, np.ndarray]], addendum: str = ""
    ) -> None:
        """Saves a list of best performances.

        output_filepath: Path or str
            path where the figure must be saved
        optim_vals: dict
            dict of losses obtained by a given optimizer.
        """
        best_performance: tp.Dict[int, tp.Any] = defaultdict(lambda: (float("inf"), "none"))
        for optim in optim_vals.keys():
            for i, l in zip(optim_vals[optim]["budget"], optim_vals[optim]["loss"]):
                if l < best_performance[i][0]:
                    best_performance[i] = (l, optim)

        with open(output_filepath, "w") as f:
            f.write(addendum)
            f.write("Best performance:\n")
            for i in best_performance.keys():
                f.write(
                    f"  budget {i}: {best_performance[i][0]} ({best_performance[i][1]}) ({output_filepath})\n"
                )

    def save(self, output_filepath: tp.PathLike) -> None:
        """Saves the xp plot

        Parameters
        ----------
        output_filepath: Path or str
            path where the figure must be saved
        """
        try:  # Let us catch errors due to too many DPIs.
            self._fig.savefig(
                str(output_filepath), bbox_extra_artists=self._overlays, bbox_inches="tight", dpi=_DPI
            )
        except ValueError as v:
            print(f"We catch {v} which means that image = too big.")
            self._fig.savefig(
                str(output_filepath), bbox_extra_artists=self._overlays, bbox_inches="tight", dpi=_DPI / 5
            )

    def __del__(self) -> None:
        plt.close(self._fig)


def split_long_title(title: str) -> str:
    """Splits a long title around the middle comma"""
    if len(title) <= 60:
        return title
    comma_indices = np.where(np.array(list(title)) == ",")[0]
    if not comma_indices.size:
        return title
    best_index = comma_indices[np.argmin(abs(comma_indices - len(title) // 2))]
    title = title[: (best_index + 1)] + "\n" + title[(best_index + 1) :]
    return title


# @contextlib.contextmanager
# def xticks_on_top() -> tp.Iterator[None]:
#     values_for_top = {'xtick.bottom': False, 'xtick.labelbottom': False,
#                       'xtick.top': True, 'xtick.labeltop': True}
#     defaults = {x: plt.rcParams[x] for x in values_for_top if x in plt.rcParams}
#     plt.rcParams.update(values_for_top)
#     yield
#     plt.rcParams.update(defaults)


class FightPlotter:
    """Creates a fight plot out of the given dataframe, by iterating over all cases with fixed category variables.

    Parameters
    ----------
    winrates_df: pd.DataFrame
        winrate data as a dataframe
    """

    def __init__(self, winrates_df: pd.DataFrame) -> None:
        # make plot
        self.winrates = winrates_df
        self._fig = plt.figure()
        self._ax = self._fig.add_subplot(111)
        max_cols = 25
        self._cax = self._ax.imshow(
            100 * np.array(self.winrates)[:, :max_cols],
            cmap=cm.seismic,
            interpolation="none",
            vmin=0,
            vmax=100,
        )
        x_names = self.winrates.columns[:max_cols]  # we plot only the 50 best
        self._ax.set_xticks(list(range(len(x_names))))
        self._ax.set_xticklabels(x_names, rotation=45, ha="right", fontsize=7)
        y_names = self.winrates.index
        self._ax.set_yticks(list(range(len(y_names))))
        self._ax.set_yticklabels(y_names, rotation=45, fontsize=7)
        divider = make_axes_locatable(self._ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        # self._fig.colorbar(im, cax=cax)
        self._fig.colorbar(self._cax, cax=cax)  # , orientation='horizontal')
        plt.tight_layout()

    @staticmethod
    def winrates_from_selection(
        df: utils.Selector,
        categories: tp.List[str],
        num_rows: int = 5,
        num_cols: int = 350,
        complete_runs_only: bool = False,
    ) -> pd.DataFrame:
        """Creates a fight plot win rate data out of the given run dataframe,
        by iterating over all cases with fixed category variables.

        Parameters
        ----------
        df: pd.DataFrame
            run data
        categories: list
            List of variables to fix for obtaining similar run conditions
        num_rows: int
            number of rows to plot (best algorithms)
        complete_runs_only: bool
            if we want a plot with only algorithms which have run on all settings
        """
        all_optimizers = list(df.unique("optimizer_name"))  # optimizers for which no run exists are not shown
        num_rows = min(num_rows, len(all_optimizers))
        # iterate on all sub cases
        victories, total = aggregate_winners(df, categories, all_optimizers)
        if complete_runs_only:
            max_num = max([int(2 * victories.loc[n, n]) for n in all_optimizers])
            new_all_optimizers = [n for n in all_optimizers if int(2 * victories.loc[n, n]) == max_num]
            if len(new_all_optimizers) > 0:
                df = df[df["optimizer_name"].isin(new_all_optimizers)]
                victories, total = aggregate_winners(df, categories, new_all_optimizers)
        # subcases = df.unique(categories)
        # for k, subcase in enumerate(subcases):  # TODO linearize this (precompute all subcases)? requires memory
        #     # print(subcase)
        #     subdf = df.select(**dict(zip(categories, subcase)))
        #     victories += _make_winners_df(subdf, all_optimizers)
        #     if k > 1000:
        #         break
        winrates = _make_sorted_winrates_df(victories)
        mean_win = winrates.mean(axis=1)
        winrates.fillna(0.5)  # unplayed
        sorted_names = winrates.index
        # number of subcases actually computed is twice self-victories
        sorted_names = ["{} ({}/{})".format(n, int(2 * victories.loc[n, n]), total) for n in sorted_names]
        num_names = len(sorted_names)
        sorted_names = [sorted_names[i] for i in range(min(num_cols, num_names))]
        data = np.array(winrates.iloc[:num_rows, : len(sorted_names)])
        # pylint: disable=anomalous-backslash-in-string
        best_names = [
            (
                f"{name} ({i+1}/{num_names}:{100 * val:2.1f}% +- {25 * np.sqrt(val*(1-val)/int(2 * victories.loc[name, name])):2.1f})"
            ).replace("Search", "")
            for i, (name, val) in enumerate(zip(mean_win.index[:num_rows], mean_win))
        ]
        return pd.DataFrame(index=best_names, columns=sorted_names, data=data)

    def save(self, *args: tp.Any, **kwargs: tp.Any) -> None:
        """Shortcut to the figure savefig method"""
        self._fig.savefig(*args, **kwargs)

    def __del__(self) -> None:
        plt.close(self._fig)


# %% positionning legends


class LegendGroup:
    """Class used to compute legend best placements.
    Each group contains at least one legend, and has a position and span (with bounds). LegendGroup are then
    responsible for providing each of its legends' position (non-overlapping)


    Parameters
    ----------
    indices: List[int]
        identifying index of each of the legends
    init_position: List[float]
        best position for each of the legends (if there was no overlapping)
    min_diff: float
        minimal distance between two legends so that they do not overlap
    """

    def __init__(self, indices: tp.List[int], init_positions: tp.List[float], min_diff: float):
        assert all(x2 - x1 == 1 for x2, x1 in zip(indices[1:], indices[:-1]))
        assert all(v2 >= v1 for v2, v1 in zip(init_positions[1:], init_positions[:-1]))
        assert len(indices) == len(init_positions)
        self.indices = indices
        self.init_positions = init_positions
        self.min_diff = min_diff
        self.position = float(np.mean(init_positions))

    def combine_with(self, other: "LegendGroup") -> "LegendGroup":
        assert self.min_diff == other.min_diff
        return LegendGroup(
            self.indices + other.indices, self.init_positions + other.init_positions, self.min_diff
        )

    def get_positions(self) -> tp.List[float]:
        first_position = self.bounds[0] + self.min_diff / 2.0
        return [first_position + k * self.min_diff for k in range(len(self.indices))]

    @property
    def bounds(self) -> tp.Tuple[float, float]:
        half_span = len(self.indices) * self.min_diff / 2.0
        return (self.position - half_span, self.position + half_span)

    def __repr__(self) -> str:
        return f"LegendGroup({self.indices}, {self.init_positions}, {self.min_diff})"


def compute_best_placements(positions: tp.List[float], min_diff: float) -> tp.List[float]:
    """Provides a list of new positions from a list of initial position, with a minimal
    distance between each position.

    Parameters
    ----------
    positions: List[float]
        best positions if minimal distance were 0.
    min_diff: float
        minimal distance allowed between two positions

    Returns
    -------
    new_positions: List[float]
        positions after taking into account the minimal distance constraint

    Note
    ----
    This function is probably not optimal, but seems a very good heuristic
    """
    assert all(v2 >= v1 for v2, v1 in zip(positions[1:], positions[:-1]))
    groups = [LegendGroup([k], [pos], min_diff) for k, pos in enumerate(positions)]
    new_groups: tp.List[LegendGroup] = []
    ready = False
    while not ready:
        ready = True
        for k in range(len(groups)):  # pylint: disable=consider-using-enumerate
            if k < len(groups) - 1 and groups[k + 1].bounds[0] < groups[k].bounds[1]:
                # groups are overlapping: create a new combined group
                # which will provide new non-overlapping positions around the mean of initial positions
                new_groups.append(groups[k].combine_with(groups[k + 1]))
                # copy the rest of the groups and start over from the first group
                new_groups.extend(groups[k + 2 :])
                groups = new_groups
                new_groups = []
                ready = False
                break
            new_groups.append(groups[k])
    new_positions = np.array(positions, copy=True)
    for group in groups:
        new_positions[group.indices] = group.get_positions()
    return new_positions.tolist()


def main() -> None:
    parser = argparse.ArgumentParser(description="Create plots from an experiment data file")
    parser.add_argument("filepath", type=str, help="filepath containing the experiment data")
    parser.add_argument(
        "--output",
        type=str,
        default=None,
        help="Output path for the CSV file (default: a folder <filename>_plots next to the data file.",
    )
    parser.add_argument(
        "--max_combsize",
        type=int,
        default=0,
        help="maximum number of parameters to fix (combinations) when creating experiment plots",
    )
    parser.add_argument(
        "--pseudotime",
        nargs="?",
        default=False,
        const=True,
        help="Plots with respect to pseudotime instead of budget",
    )
    parser.add_argument(
        "--competencemaps", type=bool, default=False, help="whether we should export only competence maps"
    )
    parser.add_argument(
        "--nomanyxp",
        type=bool,
        default=False,
        help="whether we should remove the export of detailed convergence curves",
    )
    parser.add_argument(
        "--merge-parametrization",
        action="store_true",
        help="if present, parametrization is merge into the optimizer name",
    )
    parser.add_argument(
        "--remove-suffix",
        action="store_true",
        help="if present, remove numerical suffixes in fight plots",
    )
    parser.add_argument(
        "--merge-pattern",
        type=str,
        default="",
        help="if present, optimizer name is updated according to the pattern as "
        f"an f-string. --merge-parametrization is equivalent to using --merge-pattern with {_PARAM_MERGE_PATTERN!r}",
    )
    args = parser.parse_args()
    exp_df = merge_optimizer_name_pattern(
        utils.Selector.read_csv(args.filepath),
        args.merge_pattern,
        args.merge_parametrization,
        args.remove_suffix,
    )
    # merging names
    #
    output_dir = args.output
    if output_dir is None:
        output_dir = str(Path(args.filepath).with_suffix("")) + "_plots"
    create_plots(
        exp_df,
        output_folder=output_dir,
        max_combsize=args.max_combsize if not args.competencemaps else 2,
        xpaxis="pseudotime" if args.pseudotime else "budget",
        competencemaps=args.competencemaps,
        nomanyxp=args.nomanyxp,
    )


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from . import core


def test_fishing() -> None:
    np.random.seed(17)
    func = core.OptimizeFish()
    x = np.random.rand(func.dimension)
    value = func(x)  # should not touch boundaries, so value should be < np.inf
    np.testing.assert_almost_equal(value, -0.985, decimal=2)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest
from . import core


@pytest.mark.parametrize("complex_tsp", [True, False])
def test_stsp(complex_tsp: bool) -> None:
    func = core.STSP(complex_tsp=complex_tsp)
    x = 7 * np.random.rand(func.dimension)
    value = func(x)  # should not touch boundaries, so value should be < np.inf
    assert isinstance(value, float)
    assert value < np.inf



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest
from nevergrad.optimization import optimizerlib
from nevergrad.optimization import helpers
from . import agents
from . import envs
from . import base


def test_play_environment() -> None:
    mgame = envs.DoubleOSeven(verbose=True)
    player_0 = agents.RandomAgent(mgame)
    player_1 = agents.RandomAgent(mgame)
    runner = base.EnvironmentRunner(mgame)
    rewards = runner.run(player_1=player_1, player_0=player_0)
    assert isinstance(rewards, dict)
    assert sum(rewards.values()) in [0, 1]


def test_play_single_agent_environment() -> None:
    mgame = envs.DoubleOSeven()
    game = mgame.with_agent(player_1=agents.RandomAgent(mgame)).as_single_agent()
    rand = agents.RandomAgent(game)
    runner = base.EnvironmentRunner(game)
    reward = runner.run(rand)
    assert isinstance(reward, float)
    assert reward in [0, 1]


def test_torch_agent() -> None:
    mgame = envs.DoubleOSeven()
    game = mgame.with_agent(player_1=agents.RandomAgent(mgame)).as_single_agent()
    obs = game.reset()
    agent = agents.TorchAgent.from_module_maker(game, agents.DenseNet)
    output = agent.act(obs, 0.0, False, None)
    assert output in game.action_space


def test_torch_agent_function() -> None:
    mgame = envs.DoubleOSeven()
    game = mgame.with_agent(player_1=agents.RandomAgent(mgame)).as_single_agent()
    agent = agents.TorchAgent.from_module_maker(game, agents.DenseNet)
    runner = base.EnvironmentRunner(game)
    agentfunction = agents.TorchAgentFunction(agent, runner)
    instru = agentfunction.parametrization
    args, kwargs = instru.spawn_child().set_standardized_data([0] * instru.dimension).value
    assert not args
    value = agentfunction.compute(**kwargs)
    assert value in [0, -1]  # negated reward, for minimization
    # optimization
    opt = optimizerlib.OnePlusOne(instru, budget=10)
    opt.minimize(agentfunction.compute)


def test_partial_double_seven() -> None:
    mgame = envs.DoubleOSeven()
    mgame.verbose = True
    game = mgame.with_agent(player_1=agents.RandomAgent(mgame)).as_single_agent()
    done = False
    game.reset()
    while not done:
        _, rew, done, _ = game.step(2)
    assert rew == 0


def test_torch_optimization() -> None:
    module = agents.Perceptron(input_shape=(2,), output_size=1)
    torchopt = helpers.TorchOptimizer(module.parameters(), optimizerlib.OnePlusOne)
    # pylint: disable=not-callable
    x = module.forward(agents.torch.tensor(np.array([2, 3], dtype=np.float32))).item()
    torchopt.step(3.0)
    y = module.forward(agents.torch.tensor(np.array([2, 3], dtype=np.float32))).item()
    assert x != y
    # Other instanciations
    helpers.TorchOptimizer(module.parameters(), "OnePlusOne")
    with pytest.raises(TypeError):
        helpers.TorchOptimizer(module.parameters(), 12)  # type: ignore



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import sys
import time
import random
import numbers
import warnings
import traceback
import nevergrad.common.typing as ngtp
import typing as tp
import numpy as np
from nevergrad.parametrization import parameter as p
from nevergrad.common import decorators
from nevergrad.common import errors

try:
    from ..functions.rl.agents import torch  # import includes pytorch fix
except ModuleNotFoundError:
    pass
from ..functions import base as fbase
from ..optimization import base as obase
from ..optimization.optimizerlib import (
    registry as optimizer_registry,
)  # import from optimizerlib so as to fill it
from . import execution


registry: decorators.Registry[tp.Callable[..., tp.Iterator["Experiment"]]] = decorators.Registry()


# pylint: disable=unused-argument
def _assert_singleobjective_callback(optimizer: obase.Optimizer, candidate: p.Parameter, loss: float) -> None:
    if optimizer.num_tell <= 1 and not isinstance(loss, numbers.Number):
        raise TypeError(
            f"Cannot process loss {loss} of type {type(loss)}.\n"
            "For multiobjective functions, did you forget to specify 'func.multiobjective_upper_bounds'?"
        )


class OptimizerSettings:
    """Handle for optimizer settings (name, num_workers etc)
    Optimizers can be instantiated through this class, providing the optimization space dimension.

    Note
    ----
    Eventually, this class should be moved to be directly used for defining experiments.
    """

    def __init__(
        self,
        optimizer: tp.Union[str, obase.ConfiguredOptimizer],
        budget: int,
        num_workers: int = 1,
        batch_mode: bool = True,
    ) -> None:
        self._setting_names = [x for x in locals() if x != "self"]
        if isinstance(optimizer, str):
            assert optimizer in optimizer_registry, f"{optimizer} is not registered"
        self.optimizer = optimizer
        self.budget = budget
        self.num_workers = num_workers
        self.executor = execution.MockedTimedExecutor(batch_mode)

    @property
    def name(self) -> str:
        return self.optimizer if isinstance(self.optimizer, str) else repr(self.optimizer)

    @property
    def batch_mode(self) -> bool:
        return self.executor.batch_mode

    def __repr__(self) -> str:
        return f"Experiment: {self.name}<budget={self.budget}, num_workers={self.num_workers}, batch_mode={self.batch_mode}>"

    def _get_factory(self) -> tp.Union[tp.Type[obase.Optimizer], obase.ConfiguredOptimizer]:
        return optimizer_registry[self.optimizer] if isinstance(self.optimizer, str) else self.optimizer

    @property
    def is_incoherent(self) -> bool:
        """Flags settings which are known to be impossible to process.
        Currently, this means we flag:
        - no_parallelization optimizers for num_workers > 1
        """
        # flag no_parallelization when num_workers greater than 1
        return self._get_factory().no_parallelization and bool(self.num_workers > 1)

    def instantiate(self, parametrization: p.Parameter) -> obase.Optimizer:
        """Instantiate an optimizer, providing the optimization space dimension"""
        return self._get_factory()(
            parametrization=parametrization, budget=self.budget, num_workers=self.num_workers
        )

    def get_description(self) -> tp.Dict[str, tp.Any]:
        """Returns a dictionary describing the optimizer settings"""
        descr = {x: getattr(self, x) for x in self._setting_names if x != "optimizer"}
        descr["optimizer_name"] = self.name
        return descr

    def __eq__(self, other: tp.Any) -> bool:
        if isinstance(other, self.__class__):
            for attr in self._setting_names:
                x, y = (getattr(settings, attr) for settings in [self, other])
                if x != y:
                    return False
            return True
        return False


def create_seed_generator(seed: tp.Optional[int]) -> tp.Iterator[tp.Optional[int]]:
    """Create a stream of seeds, independent from the standard random stream.
    This is designed to be used in experiment plans generators, fore reproducibility.

    Parameter
    ---------
    seed: int or None
        the initial seed

    Yields
    ------
    int or None
        potential new seeds, or None if the initial seed was None
    """
    generator = None if seed is None else np.random.RandomState(seed=seed)
    while True:
        yield None if generator is None else generator.randint(2**32, dtype=np.uint32)


class Experiment:
    """Specifies an experiment which can be run in benchmarks.

    Parameters
    ----------
    function: ExperimentFunction
        the function to run the experiment on. It must inherit from ExperimentFunction to implement
        necessary functionalities (parametrization, descriptors, evaluation_function, pseudotime etc)

    Note
    ----
    - "run" method catches error but forwards stderr so that errors are not completely hidden
    - "run" method outputs the description of the experiment, which is a set of figures/names from the functions
    settings (dimension, etc...), the optimization settings (budget, etc...) and the results (loss, etc...)
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        function: fbase.ExperimentFunction,
        optimizer: tp.Union[str, obase.ConfiguredOptimizer],
        budget: int,
        num_workers: int = 1,
        batch_mode: bool = True,
        seed: tp.Optional[int] = None,
        constraint_violation: tp.Optional[ngtp.ArrayLike] = None,
    ) -> None:
        assert isinstance(function, fbase.ExperimentFunction), (
            "All experiment functions should " "derive from ng.functions.ExperimentFunction"
        )
        assert function.dimension, "Nothing to optimize"
        self.function = function
        self.constraint_violation = constraint_violation
        self.seed = (
            seed  # depending on the inner workings of the function, the experiment may not be repeatable
        )
        self.optimsettings = OptimizerSettings(
            optimizer=optimizer, num_workers=num_workers, budget=budget, batch_mode=batch_mode
        )
        self.result = {"loss": np.nan, "elapsed_budget": np.nan, "elapsed_time": np.nan, "error": ""}
        self._optimizer: tp.Optional[
            obase.Optimizer
        ] = None  # to be able to restore stopped/checkpointed optimizer

        # make sure the random_state of the base function is created, so that spawning copy does not
        # trigger a seed for the base function, but only for the copied function
        self.function.parametrization.random_state  # pylint: disable=pointless-statement

    def __repr__(self) -> str:
        return f"Experiment: {self.optimsettings} (dim={self.function.dimension}, param={self.function.parametrization}) on {self.function} with seed {self.seed}"

    @property
    def is_incoherent(self) -> bool:
        """Flags settings which are known to be impossible to process.
        Currently, this means we flag:
        - no_parallelization optimizers for num_workers > 1
        """
        return self.optimsettings.is_incoherent

    def run(self) -> tp.Dict[str, tp.Any]:
        """Run an experiment with the provided settings

        Returns
        -------
        dict
            A dict containing all the information about the experiments (optimizer/function settings + results)

        Note
        ----
        This function catches error (but forwards stderr). It fills up the "error" ("" if no error, else the error name),
        "loss", "elapsed_time" and "elapsed_budget" of the experiment.
        """
        try:
            self._run_with_error()
        except (errors.ExperimentFunctionCopyError, errors.UnsupportedExperiment) as ex:
            raise ex
        except Exception as e:  # pylint: disable=broad-except
            # print the case and the traceback
            self.result["error"] = e.__class__.__name__
            print(f"Error when applying {self}:", file=sys.stderr)
            traceback.print_exc()
            print("\n", file=sys.stderr)
        return self.get_description()

    def _log_results(self, pfunc: fbase.ExperimentFunction, t0: float, num_calls: int) -> None:
        """Internal method for logging results before handling the error"""
        self.result["elapsed_time"] = time.time() - t0
        self.result["pseudotime"] = self.optimsettings.executor.time
        # make a final evaluation with oracle (no noise, but function may still be stochastic)
        opt = self._optimizer
        assert opt is not None
        # ExperimentFunction can directly override this evaluation function if need be
        # (pareto_front returns only the recommendation in singleobjective)
        self.result["num_objectives"] = opt.num_objectives
        self.result["loss"] = pfunc.evaluation_function(*opt.pareto_front())
        if (
            self.constraint_violation
            and np.sum([f(opt.recommend().value) for f in self.constraint_violation]) > 0  # type: ignore
            or len(self.function.parametrization._constraint_checkers) > 0
            and not opt.recommend().satisfies_constraints(pfunc.parametrization)
        ):
            self.result["loss"] += 1e9  # type: ignore
        self.result["elapsed_budget"] = num_calls
        if num_calls > self.optimsettings.budget:
            raise RuntimeError(
                f"Too much elapsed budget {num_calls} for {self.optimsettings.name} on {self.function}"
            )
        self.result.update({f"info/{x}": y for x, y in opt._info().items()})  # Add optimizer info for debug

    def _run_with_error(self, callbacks: tp.Optional[tp.Dict[str, obase._OptimCallBack]] = None) -> None:
        """Run an experiment with the provided artificial function and optimizer

        Parameter
        ---------
        callbacks: dict
            a dictionary of callbacks to register on the optimizer with key "ask" and/or "tell" (see base Optimizer class).
            This is only for easier debugging.
        """
        if self.seed is not None and self._optimizer is None:
            # Note: when resuming a job (if optimizer is not None), seeding is pointless (reproducibility is lost)
            np.random.seed(
                self.seed
            )  # seeds both functions and parametrization (for which random state init is lazy)
            random.seed(self.seed)
            torch.manual_seed(self.seed)
        pfunc = self.function.copy()
        # check constraints are propagated
        assert len(pfunc.parametrization._constraint_checkers) == len(
            self.function.parametrization._constraint_checkers
        )
        # optimizer instantiation can be slow and is done only here to make xp iterators very fast
        if self._optimizer is None:
            self._optimizer = self.optimsettings.instantiate(parametrization=pfunc.parametrization)
            if pfunc.multiobjective_upper_bounds is not None:
                self._optimizer.tell(p.MultiobjectiveReference(), pfunc.multiobjective_upper_bounds)
            else:
                self._optimizer.register_callback("tell", _assert_singleobjective_callback)
        if callbacks is not None:
            for name, func in callbacks.items():
                self._optimizer.register_callback(name, func)
        assert self._optimizer.budget is not None, "A budget must be provided"
        t0 = time.time()
        executor = self.optimsettings.executor
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", category=errors.InefficientSettingsWarning
            )  # benchmark do not need to be efficient
            try:
                # call the actual Optimizer.minimize method because overloaded versions could alter the worklflow
                # and provide unfair comparisons  (especially for parallelized settings)
                obase.Optimizer.minimize(
                    self._optimizer,
                    pfunc,
                    batch_mode=executor.batch_mode,
                    executor=executor,
                    constraint_violation=self.constraint_violation,
                )
            except Exception as e:  # pylint: disable=broad-except
                self._log_results(pfunc, t0, self._optimizer.num_ask)
                raise e
        self._log_results(pfunc, t0, self._optimizer.num_ask)

    def get_description(self) -> tp.Dict[str, tp.Union[str, float, bool]]:
        """Return the description of the experiment, as a dict.
        "run" must be called beforehand in order to have non-nan values for the loss.
        """
        summary = dict(self.result, seed=-1 if self.seed is None else self.seed)
        summary.update(self.function.descriptors)
        summary.update(self.optimsettings.get_description())
        return summary  # type: ignore

    def __eq__(self, other: tp.Any) -> bool:
        if not isinstance(other, Experiment):
            return False
        same_seed = other.seed is None if self.seed is None else other.seed == self.seed
        return (
            same_seed
            and self.function.equivalent_to(other.function)
            and self.optimsettings == other.optimsettings
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
import numpy as np
from nevergrad import optimizers
from nevergrad.optimization.base import ConfiguredOptimizer
from nevergrad.optimization import experimentalvariants  # pylint: disable=unused-import
from nevergrad.functions import ArtificialFunction
from .xpbase import registry
from .xpbase import create_seed_generator
from .xpbase import Experiment

# pylint: disable=stop-iteration-return, too-many-nested-blocks


@registry.register
def basic(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """Test settings"""
    seedg = create_seed_generator(seed)
    function = ArtificialFunction(name="sphere", block_dimension=2, noise_level=1)
    np.random.seed(seed)  # seed before initializing the function!
    # initialization uses randomness
    function.transform_var._initialize()
    return iter([Experiment(function, optimizer="OnePlusOne", num_workers=2, budget=4, seed=next(seedg))])


@registry.register
def repeated_basic(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """Test settings"""
    seedg = create_seed_generator(seed)
    function = ArtificialFunction(name="sphere", block_dimension=2, noise_level=1)
    optims: tp.List[tp.Union[str, ConfiguredOptimizer]] = ["OnePlusOne", optimizers.DifferentialEvolution()]
    for _ in range(5):
        for optim in optims:
            yield Experiment(function, optimizer=optim, num_workers=2, budget=4, seed=next(seedg))


@registry.register
def illcond(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """All optimizers on ill cond problems"""
    seedg = create_seed_generator(seed)
    for budget in [500, 1000, 2000, 4000]:
        for optim in ["SQP", "DE", "CMA", "PSO", "RotationInvariantDE", "NelderMead"]:
            for rotation in [True, False]:
                for name in ["ellipsoid", "cigar"]:
                    function = ArtificialFunction(name=name, rotation=rotation, block_dimension=100)
                    yield Experiment(function, optim, budget=budget, seed=next(seedg))


@registry.register
def compabasedillcond(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """All optimizers on ill cond problems"""
    seedg = create_seed_generator(seed)
    for budget in [500, 1000, 2000, 4000, 8000]:
        for optim in [
            "DE",
            "CMA",
            "PSO",
            "BPRotationInvariantDE",
            "RotationInvariantDE",
            "AlmostRotationInvariantDE",
            "AlmostRotationInvariantDEAndBigPop",
        ]:
            for rotation in [True, False]:
                for name in ["ellipsoid", "cigar"]:
                    function = ArtificialFunction(name=name, rotation=rotation, block_dimension=30)
                    yield Experiment(function, optim, budget=budget, seed=next(seedg))


@registry.register
def noise(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """All optimizers on ill cond problems"""
    seedg = create_seed_generator(seed)
    optims = sorted(
        x
        for x, y in optimizers.registry.items()
        if ("TBPSA" in x or "ois" in x or "CMA" in x or "epea" in x) and "iscr" not in x
    )
    for budget in [500, 1000, 2000, 4000, 8000, 16000, 32000, 64000, 128000]:
        for optim in optims:
            for rotation in [True, False]:
                for name in ["sphere", "cigar", "sphere4"]:
                    function = ArtificialFunction(
                        name=name, rotation=rotation, block_dimension=20, noise_level=10
                    )
                    yield Experiment(function, optim, budget=budget, seed=next(seedg))


@registry.register
def dim10_smallbudget(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    # prepare list of parameters to sweep for independent variables
    seedg = create_seed_generator(seed)
    names = ["sphere"]
    optims = sorted(
        x for x, y in optimizers.registry.items() if y.one_shot and "arg" not in x and "mal" not in x
    )
    functions = [
        ArtificialFunction(
            name, block_dimension=bd, num_blocks=n_blocks, useless_variables=bd * uv_factor * n_blocks
        )
        for name in names
        for bd in [10]
        for uv_factor in [0]
        for n_blocks in [1]
    ]
    for func in functions:
        for optim in optims:
            for budget in [4, 8, 16, 32]:
                yield Experiment(func, optim, budget=budget, num_workers=1, seed=next(seedg))


@registry.register
# 2 variables matter - Scrambled Hammersley rules.
def dim10_select_two_features(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    # prepare list of parameters to sweep for independent variables
    seedg = create_seed_generator(seed)
    names = ["sphere"]
    optims = sorted(
        x for x, y in optimizers.registry.items() if y.one_shot and "arg" not in x and "mal" not in x
    )
    functions = [
        ArtificialFunction(
            name, block_dimension=bd, num_blocks=n_blocks, useless_variables=bd * uv_factor * n_blocks
        )
        for name in names
        for bd in [2]
        for uv_factor in [5]
        for n_blocks in [1]
    ]
    for func in functions:
        for optim in optims:
            for budget in [4, 8, 16, 32]:
                yield Experiment(func, optim, budget=budget, num_workers=1, seed=next(seedg))


@registry.register
def dim10_select_one_feature(
    seed: tp.Optional[int] = None,
) -> tp.Iterator[Experiment]:  # One and only one variable matters - LHS wins.
    # prepare list of parameters to sweep for independent variables
    seedg = create_seed_generator(seed)
    names = ["sphere"]
    optims = sorted(
        x for x, y in optimizers.registry.items() if y.one_shot and "arg" not in x and "mal" not in x
    )
    functions = [
        ArtificialFunction(
            name, block_dimension=bd, num_blocks=n_blocks, useless_variables=bd * uv_factor * n_blocks
        )
        for name in names
        for bd in [1]
        for uv_factor in [10]
        for n_blocks in [1]
    ]
    for func in functions:
        for optim in optims:
            for budget in [8, 10, 12, 14, 16, 18, 20]:
                yield Experiment(func, optim, budget=budget, num_workers=1, seed=next(seedg))


@registry.register
def doe_dim4(
    seed: tp.Optional[int] = None,
) -> tp.Iterator[Experiment]:  # Here, QR performs best, then Random, then LHS, then Cauchy.
    # prepare list of parameters to sweep for independent variables
    seedg = create_seed_generator(seed)
    names = ["sphere"]  # n for n in ArtificialFunction.list_sorted_function_names() if "sphere" in n]
    optims = sorted(
        x for x, y in optimizers.registry.items() if y.one_shot and "arg" not in x and "mal" not in x
    )
    functions = [
        ArtificialFunction(
            name, block_dimension=bd, num_blocks=n_blocks, useless_variables=bd * uv_factor * n_blocks
        )
        for name in names
        for bd in [4]
        for uv_factor in [0]
        for n_blocks in [1]
    ]
    for func in functions:
        for optim in optims:
            for budget in [30, 100, 3000, 10000]:
                yield Experiment(func, optim, budget=budget, num_workers=1, seed=next(seedg))


@registry.register
def oneshot4(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    # General experiment comparing one-shot optimizers, excluding those with "large" or "small"
    # in the name.
    seedg = create_seed_generator(seed)
    names = ["sphere", "cigar", "ellipsoid", "rosenbrock", "rastrigin"]
    optims = sorted(
        x for x, y in optimizers.registry.items() if y.one_shot and "arg" not in x and "mal" not in x
    )
    functions = [
        ArtificialFunction(
            name, block_dimension=bd, num_blocks=n_blocks, useless_variables=bd * uv_factor * n_blocks
        )
        for name in names
        for bd in [1, 4, 20]
        for uv_factor in [0, 10]
        for n_blocks in [1]
    ]
    for func in functions:
        for optim in optims:
            for budget in [30, 100, 3000]:
                yield Experiment(func, optim, budget=budget, num_workers=1, seed=next(seedg))


@registry.register
def oneshot3(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    # General experiment comparing one-shot optimizers, excluding those with "large" or "small"
    # in the name.
    seedg = create_seed_generator(seed)
    names = ["sphere", "altcigar", "cigar", "ellipsoid", "rosenbrock", "rastrigin", "altellipsoid"]
    optims = sorted(
        x for x, y in optimizers.registry.items() if y.one_shot and "arg" not in x and "mal" not in x
    )
    functions = [ArtificialFunction(name, block_dimension=bd) for name in names for bd in [4, 20]]
    for func in functions:
        for optim in optims:
            for budget in [30, 60, 100]:
                yield Experiment(func, optim, budget=budget, num_workers=1, seed=next(seedg))


@registry.register
def oneshot2(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    # Experiment comparing one-shot optimizers in the context of useless vars vs critical vars.
    seedg = create_seed_generator(seed)
    names = ["sphere", "altcigar", "cigar", "ellipsoid", "rosenbrock", "rastrigin", "altellipsoid"]
    optims = sorted(
        x for x, y in optimizers.registry.items() if y.one_shot and "arg" not in x and "mal" not in x
    )
    functions = [
        ArtificialFunction(name, block_dimension=2, num_blocks=1, useless_variables=20) for name in names
    ]
    for func in functions:
        for optim in optims:
            for budget in [30, 60, 100]:
                yield Experiment(func, optim, budget=budget, num_workers=1, seed=next(seedg))


@registry.register
def oneshot1(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """Comparing one-shot optimizers as initializers for Bayesian Optimization."""
    seedg = create_seed_generator(seed)
    for budget in [25, 31, 37, 43, 50, 60]:  # , 4000, 8000, 16000, 32000]:
        for optim in sorted(x for x, y in optimizers.registry.items() if "BO" in x):
            for rotation in [False]:
                for d in [20]:
                    for name in ["sphere", "cigar", "hm", "ellipsoid"]:  # , "hm"]:
                        for u in [0]:
                            function = ArtificialFunction(
                                name=name,
                                rotation=rotation,
                                block_dimension=d,
                                useless_variables=d * u,
                                translation_factor=1.0,
                            )
                            yield Experiment(function, optim, budget=budget, seed=next(seedg))


@registry.register
def metanoise(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    seedg = create_seed_generator(seed)
    optims = ["NoisyBandit", "TBPSA", "NaiveTBPSA"]
    for budget in [15, 31, 62, 125, 250, 500, 1000, 2000, 4000, 8000]:
        for optim in optims:
            for noise_dissymmetry in [False, True]:
                function = ArtificialFunction(
                    name="sphere",
                    rotation=True,
                    block_dimension=1,
                    noise_level=10,
                    noise_dissymmetry=noise_dissymmetry,
                    translation_factor=10.0,
                )
                yield Experiment(function, optim, budget=budget, seed=next(seedg))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from . import core


def test_unit_commitment_p1() -> None:
    np.random.seed(0)
    T = 10
    N = 5
    func = core.UnitCommitmentProblem(problem_name="semi-continuous", num_timepoints=T, num_generators=N)
    op_out = np.ones((N, T))
    op_states = np.ones((N, T))
    value = func.function(operational_output=op_out, operational_states=op_states)
    assert isinstance(value, float)
    assert np.allclose([value], [38721960.61493097], rtol=1e-04, atol=1e-05)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import scipy
import scipy.signal
import scipy.stats
import time
import copy
import numpy as np
import itertools
from joblib import Parallel, delayed  # type: ignore

# import matplotlib as mpl
# import matplotlib.pyplot as plt
from collections import defaultdict
import nevergrad as ng

# pylint: skip-file

default_budget = 3000  # centiseconds
methods = {}
metrics = {}

# A few helper functions.
def normalize(x):
    for i in range(len(x)):
        normalization = np.sqrt(np.sum(x[i] ** 2.0))
        if normalization > 0.0:
            x[i] = x[i] / normalization
        else:
            x[i] = np.random.randn(np.prod(x[i].shape)).reshape(x[i].shape)
            x[i] = x[i] / np.sqrt(np.sum(x[i] ** 2.0))
    return x


def convo(x, k):
    if k is None:
        return x
    return scipy.ndimage.gaussian_filter(x, sigma=list(k) + [0.0] * (len(x.shape) - len(k)))


# Our well distributed point configurations.
def pure_random(n, shape, conv=None):
    return normalize([np.random.randn(*shape) for i in range(n)])


def antithetic_pm(n, shape, conv=None):
    m = n // 2
    x = [np.random.randn(*shape) for i in range(m)]
    x = normalize(x)
    x = x + [-xi for xi in x]
    if len(x) < n:
        x = x + [np.random.randn(*shape)]
    return np.array(x)


def antithetic_order(n, shape, axis=-1, also_sym=False, conv=None):
    x = []
    s = shape[axis]
    indices = [slice(0, s, 1) for s in shape]
    indices_sym = [slice(0, s, 1) for s in shape]
    while len(x) < n:
        icx = normalize([np.random.randn(*shape)])[0]
        for p in itertools.permutations(range(s)):
            if len(x) < n:
                indices[axis] = p
                cx = copy.deepcopy(icx)[tuple(indices)]
                x = x + [cx]
                if also_sym:
                    order = list(itertools.product((False, True), repeat=s))
                    np.random.shuffle(order)
                    for ordering in order:  # Ordering is a list of bool
                        if any(ordering) and len(x) < n:
                            scx = copy.deepcopy(cx)
                            for o in [i for i, o in enumerate(ordering) if o]:  # we must symetrize o
                                indices_sym[axis] = o
                                scx[tuple(indices_sym)] = -scx[tuple(indices_sym)]
                            x = x + [scx]
    return x


def antithetic_order_and_sign(n, shape, axis=-1, conv=None):
    return antithetic_order(n, shape, axis, also_sym=True)


def greedy_dispersion(n, shape, budget=default_budget, conv=None):
    x = normalize([np.random.randn(*shape)])

    for i in range(n - 1):
        bigdist = -1
        t0 = time.time()
        while time.time() < t0 + 0.01 * budget / n:
            y = normalize([np.random.randn(*shape)])[0]
            dist = min(np.linalg.norm(convo(y, conv) - convo(x[i], conv)) for i in range(len(x)))
            if dist > bigdist:
                bigdist = dist
                newy = y
        x += [newy]
    return x


def dispersion(n, shape, budget=default_budget, conv=None):
    x = greedy_dispersion(n, shape, budget / 2, conv=conv)
    t0 = time.time()
    num = n
    while time.time() < t0 + 0.01 * budget / 2:
        num = num + 1
        for j in range(len(x)):
            bigdist = -1
            for idx in range(2 * num):
                if idx > 0:
                    y = normalize([np.random.randn(*shape)])[0]
                else:
                    y = x[j]
                convoy = convo(y, conv)
                dist = min(np.linalg.norm(convoy - convo(x[i], conv)) for i in range(len(x)) if i != j)
                if dist > bigdist:
                    x[j] = y
                    bigdist = dist
    return x


def dispersion_with_conv(n, shape, budget=default_budget):
    return dispersion(n, shape, budget=budget, conv=[8, 8])


def greedy_dispersion_with_conv(n, shape, budget=default_budget):
    return greedy_dispersion(n, shape, budget=budget, conv=[8, 8])


def dispersion_with_big_conv(n, shape, budget=default_budget):
    return dispersion(n, shape, budget=budget, conv=[24, 24])


def greedy_dispersion_with_big_conv(n, shape, budget=default_budget):
    return greedy_dispersion(n, shape, budget=budget, conv=[24, 24])


def dispersion_with_mini_conv(n, shape, budget=default_budget):
    return dispersion(n, shape, budget=budget, conv=[2, 2])


def greedy_dispersion_with_mini_conv(n, shape, budget=default_budget):
    return greedy_dispersion(n, shape, budget=budget, conv=[2, 2])


def block_symmetry(n, shape, num_blocks=None):
    x = []
    if num_blocks is None:
        num_blocks = [4, 4]
    for pindex in range(n):
        # print(f"block symmetry {pindex}/{n}")
        newx = normalize([np.random.randn(*shape)])[0]
        s = np.prod(num_blocks)
        num_blocks = num_blocks + ([1] * (len(shape) - len(num_blocks)))
        order = list(itertools.product((False, True), repeat=s))
        np.random.shuffle(order)
        ranges = [list(range(n)) for n in num_blocks]
        for o in order:  # e.g. o is a list of 64 bool if num_blocks is 8x8
            # print(f"We have the boolean vector {o}")
            tentativex = copy.deepcopy(newx)
            for i, multi_index in enumerate(itertools.product(*ranges)):
                if o[i]:  # This multi-index corresponds to a symmetry.
                    # print("our multi-index is ", multi_index)
                    slices = [[]] * len(shape)
                    for c, p in enumerate(multi_index):
                        assert p >= 0
                        assert p < num_blocks[c]
                        a = p * shape[c] // num_blocks[c]
                        b = min((p + 1) * shape[c] // num_blocks[c], shape[c])
                        slices[c] = slice(a, b)
                    slices = tuple(slices)
                    # print(slices)
                    tentativex[slices] = -tentativex[slices]
            if len(x) >= n:
                return x
            x += [tentativex]


def big_block_symmetry(n, shape):
    return block_symmetry(n, shape, num_blocks=[2, 2])


def covering(n, shape, budget=default_budget, conv=None):
    x = greedy_dispersion_with_conv(n, shape, budget / 2, conv)
    mindists = []
    c = 0.01
    previous_score = float("inf")
    num = 0
    t0 = time.time()
    while time.time() < t0 + 0.01 * budget / 2:
        num = num + 1
        t = normalize([np.random.randn(*shape)])[0]
        convt = convo(t, conv)
        mindist = float("inf")
        for k in range(len(x)):
            dist = np.linalg.norm(convt - convo(x[k], conv))
            if dist < mindist:
                mindist = dist
                index = k
        mindists += [mindist]
        if len(mindists) % 2000 == 0:
            score = np.sum(mindists[-2000:]) / len(mindists[-2000:])
            c *= 2 if score < previous_score else 0.5
            previous_score = score
            # print(score, c)
        x[index] = normalize([x[index] + (c / (35 + n + np.sqrt(num))) * (t - x[index])])[0]
    # print("covering:", scipy.ndimage.gaussian_filter(mindists, budget / 3, mode='reflect')[::len(mindists) // 7])
    return x


def covering_conv(n, shape, budget=default_budget):
    return covering(n, shape, budget, conv=[8, 8])


def covering_mini_conv(n, shape, budget=default_budget):
    return covering(n, shape, budget, conv=[2, 2])


def get_class(x, num_blocks, just_max):
    shape = x.shape
    split_volume = len(num_blocks)
    num_blocks = num_blocks + [1] * (len(shape) - len(num_blocks))
    ranges = [list(range(n)) for n in num_blocks]
    result = []
    for _, multi_index in enumerate(itertools.product(*ranges)):
        slices = [[]] * len(shape)
        for c, p in enumerate(multi_index):
            assert p >= 0
            assert p < num_blocks[c]
            a = p * shape[c] // num_blocks[c]
            b = min((p + 1) * shape[c] // num_blocks[c], shape[c])
            slices[c] = slice(a, b)
        slices = tuple(slices)
        if just_max:
            result = result + [
                list(np.argsort((np.sum(x[slices], tuple(range(split_volume)))).flatten()))[-1]
            ]
        else:
            result = result + list(np.argsort((np.sum(x[slices], tuple(range(split_volume)))).flatten()))
    return hash(str(result))


def jittered(n, shape, num_blocks=None, just_max=False):
    if num_blocks is None:
        num_blocs = [2, 2]
    hash_to_set = defaultdict(list)
    for i in range(int(np.sqrt(n)) * n):
        x = normalize([np.random.randn(*shape)])[0]
        hash_to_set[get_class(x, num_blocks, just_max)] += [x]
    min_num = 10000000
    max_num = -1
    while True:
        for k in hash_to_set.keys():
            min_num = min(min_num, len(hash_to_set[k]))
            max_num = max(max_num, len(hash_to_set[k]))
        # print(f"min={min_num}, max={max_num}")
        if min_num < n / len(hash_to_set.keys()):
            x = normalize([np.random.randn(*shape)])[0]
            hash_to_set[get_class(x, num_blocks, just_max)] += [x]
        else:
            break
    x = []
    while len(x) < n:
        num = max(1, (n - len(x)) // len(hash_to_set))
        for k in hash_to_set.keys():
            if len(x) < n:
                # print(f"Adding {num} in {k}...")
                x += hash_to_set[k][:num]
            hash_to_set[k] = hash_to_set[k][num:]
    assert len(x) == n
    # print(f"Jittered used {len(hash_to_set)} classes")
    return x


def reduced_jittered(n, shape):
    return jittered(n, shape, [2, 2], just_max=True)


def covering_big_conv(n, shape, budget=default_budget):
    return covering(n, shape, budget, [24, 24])


def lhs(n, shape):
    num = np.prod(shape)
    x = np.zeros([n, num])
    for i in range(num):
        xb = (1.0 / n) * np.random.rand(n)
        xplus = np.linspace(0, n - 1, n) / n
        np.random.shuffle(xplus)
        x[:, i] = scipy.stats.norm.ppf(xb + xplus)
    thex = []
    for i in range(n):
        thex += normalize([x[i].reshape(*shape)])
    assert len(thex) == n
    assert thex[0].shape == tuple(shape), f" we get {x[0].shape} instead of {tuple(shape)}"
    return thex


# list_for_drawing = [
#     "lhs",
#     "reduced_jittered",
#     "jittered",
#     "big_block_symmetry",
#     "block_symmetry",
#     "greedy_dispersion",
#     "dispersion",
#     "pure_random",
#     "antithetic_pm",
#     "dispersion",
#     "antithetic_order",
#     "antithetic_order_and_sign",
#     "covering_conv",
#     "covering",
# ]
#
#
# def show_points(x, k):
#     plt.clf()
#     # fig = plt.figure()
#     ##ax = fig.gca(projection='3d')
#     # ax = fig.add_subplot(projection='3d')
#     # ax.set_aspect("equal")
#     # ax.view_init(elev=0., azim=0.)
#
#     def make_ball(a, b, c, r, col="r"):
#         # u, v = np.mgrid[0:2*np.pi:20j, 0:np.pi:10j]
#         # x = 0.5 + 0.5 * (a+r*np.cos(u)*np.sin(v))
#         # y = 0.5 + 0.5 * (b+r*np.sin(u)*np.sin(v))
#         # z = 0.5 + 0.5 * (c+r*np.cos(v))
#         # ax.plot_wireframe(x, y, z, color=col, zorder=z)
#         u = np.linspace(0, 2 * 3.14159, 40)
#         x = 0.5 + 0.5 * (a + r * np.cos(u))
#         y = 0.5 + 0.5 * (b + r * np.sin(u))
#         plt.plot(x, y, color=col)
#         plt.plot(
#             [0.5, 0.5 + 0.5 * a], [0.5, 0.5 + 0.5 * b], color=col, linestyle=("dashed" if c > 0 else "solid")
#         )
#
#     for i in range(len(x)):
#         if x[i][2] > 0:
#             make_ball(x[i][0], x[i][1], x[i][2], 0.05 * (3 / (3 + x[i][2])), "g")
#     make_ball(0, 0, 0, 1, "r")
#     for i in range(len(x)):
#         if x[i][2] <= 0:
#             make_ball(x[i][0], x[i][1], x[i][2], 0.05 * (3 / (3 + x[i][2])), "b")
#     # plt.show()
#     plt.tight_layout()
#     plt.savefig(f"sphere3d_{k}.png")


# Let us play with metrics.
def metric_half(x, budget=default_budget, conv=None):
    shape = x[0].shape
    t0 = time.time()
    xconv = np.array([convo(x_, conv).flatten() for x_ in x])
    scores = []
    while time.time() < t0 + 0.01 * budget:
        y = convo(normalize([np.random.randn(*shape)])[0], conv).flatten()
        scores += [np.average(np.matmul(xconv, y) > 0.0)]
    return np.average((np.array(scores) - 0.5) ** 2)


def metric_half_conv(x, budget=default_budget):
    return metric_half(x, budget, conv=[8, 8])


def metric_cap(x, budget=default_budget, conv=None):
    shape = x[0].shape
    t0 = time.time()
    c = 1.0 / np.sqrt(len(x[0].flatten()))
    xconv = np.array(normalize([convo(x_, conv).flatten() for x_ in x]))
    scores = []
    while time.time() < t0 + 0.01 * budget:
        y = convo(normalize([np.random.randn(*shape)])[0], conv).flatten()
        scores += [np.average(np.matmul(xconv, y) > c)]
        scores += [np.average(np.matmul(xconv, y) < -c)]
    return np.std(np.array(scores))


def metric_cap_conv(x, budget=default_budget):
    return metric_cap(x, budget, conv=[8, 8])


def metric_pack_avg(x, budget=default_budget, conv=None):
    shape = x[0].shape
    xconv = np.array(normalize([convo(x_, conv).flatten() for x_ in x]))
    scores = np.matmul(xconv, xconv.transpose())
    for i in range(len(scores)):
        assert 0.99 < scores[i, i] < 1.01
        scores[i, i] = 0
    scores = scores.flatten()
    assert len(scores) == len(x) ** 2
    return np.average(scores)


def metric_pack_avg_conv(x, budget=default_budget):
    return metric_pack_avg(x, budget=default_budget, conv=[8, 8])


def metric_pack(x, budget=default_budget, conv=None):
    shape = x[0].shape
    xconv = np.array(normalize([convo(x_, conv).flatten() for x_ in x]))
    scores = np.matmul(xconv, xconv.transpose())
    for i in range(len(scores)):
        assert 0.99 < scores[i, i] < 1.01, "we get score " + str(scores[i, i])
        scores[i, i] = 0
    scores = scores.flatten()
    assert len(scores) == len(x) ** 2
    return max(scores)


def metric_pack_conv(x, budget=default_budget):
    return metric_pack(x, budget=default_budget, conv=[8, 8])


list_of_methods = [
    "rs_ng_TwoPointsDE",
    "rs_ng_DE",
    "rs_ng_PSO",
    "rs_ng_OnePlusOne",
    "rs_ng_DiagonalCMA",
    "lhs",
    "reduced_jittered",
    "jittered",
    "big_block_symmetry",
    "block_symmetry",
    "greedy_dispersion",
    "dispersion",
    "pure_random",
    "antithetic_pm",
    "dispersion",
    "antithetic_order",
    "antithetic_order_and_sign",
    "dispersion_with_conv",
    "dispersion_with_big_conv",
    "greedy_dispersion_with_big_conv",
    "dispersion_with_mini_conv",
    "greedy_dispersion_with_mini_conv",
    "covering",
    "covering_conv",
    "covering_mini_conv",
    "rs_metric",
    "rs_metric_mhc",
    "rs_metric_pack",
    "rs_metric_pa",
    "rs_metric_pc",
    "rs_metric_pac",
    "rs_metric_cap",
    "rs_metric_cc",
    "rs_metric_all",
]
list_metrics = [
    "metric_half",
    "metric_half_conv",
    "metric_pack",
    "metric_pack_conv",
    "metric_pack_avg",
    "metric_pack_avg_conv",
    "metric_cap",
    "metric_cap_conv",
]
for u in list_metrics:
    metrics[u] = eval(u)


def rs_metric(n, shape, budget=default_budget, k="metric_half", ngtool=None):
    t0 = time.time()
    bestm = float("inf")
    if ngtool is not None:
        opt = ng.optimizers.registry[ngtool](
            ng.p.Array(shape=tuple([n] + list(shape))), budget=10000000000000
        )
    while time.time() < t0 + 0.01 * budget:
        if ngtool is None:
            x = pure_random(n, shape)
        else:
            candidate = opt.ask()
            x = list(candidate.value)
            assert len(x) == n
            x = normalize(x)
        # m = eval(f"{k}(x, budget={budget/max(10,np.sqrt(budget/100))})")
        if k == "all":
            m = np.sum(
                [
                    metrics[k2](x, (budget / len(list_metrics)) / max(10, np.sqrt(budget / 100)))
                    for k2 in list_metrics
                ]
            )
        else:
            m = metrics[k](x, budget / max(10, np.sqrt(budget / 100)))
        if ngtool is not None:
            print("ng gets ", m)
            opt.tell(candidate, m)
        if m < bestm:
            bestm = m
            bestx = x
    return bestx


def rs_metric_mhc(n, shape, budget=default_budget):
    return rs_metric(n, shape, budget, k="metric_half_conv")


def rs_metric_cap(n, shape, budget=default_budget):
    return rs_metric(n, shape, budget, k="metric_cap")


def rs_metric_cc(n, shape, budget=default_budget):
    return rs_metric(n, shape, budget, k="metric_cap_conv")


def rs_metric_pack(n, shape, budget=default_budget):
    return rs_metric(n, shape, budget, k="metric_pack")


def rs_metric_pa(n, shape, budget=default_budget):
    return rs_metric(n, shape, budget, k="metric_pack_avg")


def rs_metric_pc(n, shape, budget=default_budget):
    return rs_metric(n, shape, budget, k="metric_pack_conv")


def rs_metric_pac(n, shape, budget=default_budget):
    return rs_metric(n, shape, budget, k="metric_pack_avg_conv")


def rs_metric_all(n, shape, budget=default_budget):
    return rs_metric(n, shape, budget, k="all")


def rs_ng_TwoPointsDE(n, shape, budget=default_budget):
    return rs_metric(n, shape, budget, k="all", ngtool="TwoPointsDE")


def rs_ng_DE(n, shape, budget=default_budget):
    return rs_metric(n, shape, budget, k="all", ngtool="DE")


def rs_ng_PSO(n, shape, budget=default_budget):
    return rs_metric(n, shape, budget, k="all", ngtool="PSO")


def rs_ng_OnePlusOne(n, shape, budget=default_budget):
    return rs_metric(n, shape, budget, k="all", ngtool="OnePlusOne")


def rs_ng_DiagonalCMA(n, shape, budget=default_budget):
    return rs_metric(n, shape, budget, k="all", ngtool="DiagonalCMA")


data = defaultdict(lambda: defaultdict(list))  # type: ignore


# def do_plot(tit, values):
#     plt.clf()
#     plt.title(tit.replace("_", " "))
#     x = np.cos(np.linspace(0.0, 2 * 3.14159, 20))
#     y = np.sin(np.linspace(0.0, 2 * 3.14159, 20))
#     for i, v in enumerate(sorted(values.keys(), key=lambda k: np.average(values[k]))):
#         print(f"context {tit}, {v} ==> {values[v]}")
#         plt.plot([i + r for r in x], [np.average(values[v]) + r * np.std(values[v]) for r in y])
#         plt.text(i, np.average(values[v]) + np.std(values[v]), f"{v}", rotation=30)
#         if i > 0:
#             plt.savefig(
#                 f"comparison_{tit}_time{default_budget}.png".replace(" ", "_")
#                 .replace("]", "_")
#                 .replace("[", "_")
#             )
#
#
# def heatmap(y, x, table, name):
#     for cn in ["viridis", "plasma", "inferno", "magma", "cividis"]:
#         print(f"Creating a heatmap with name {name}: {table}")
#         plt.clf()
#         fig, ax = plt.subplots()
#         tab = copy.deepcopy(table)
#
#         for j in range(len(tab[0, :])):
#             for i in range(len(tab)):
#                 tab[i, j] = np.average(table[:, j] < table[i, j])
#         print(tab)
#         im = ax.imshow(tab, aspect="auto", cmap=mpl.colormaps[cn])
#
#         # Show all ticks and label them with the respective list entries
#         ax.set_xticks(np.arange(len(x)), labels=x)
#         ax.set_yticks(np.arange(len(y)), labels=y)
#
#         # Rotate the tick labels and set their alignment.
#         plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")
#
#         # Loop over data dimensions and create text annotations.
#         for i in range(len(y)):
#             for j in range(len(x)):
#                 text = ax.text(j, i, str(tab[i, j])[:4], ha="center", va="center", color="k")
#
#         fig.tight_layout()
#         plt.savefig(f"TIME{default_budget}_cmap{cn}" + name)
#
#
# def create_statistics(n, shape, list_of_methods, list_of_metrics, num=1):
#     for _ in range(num):
#         for idx, method in enumerate(list_of_methods):
#             print(f" {idx}/{len(list_of_methods)}: {method}")
#             x = eval(f"{method}(n, shape)")
#             for k in list_of_metrics:
#                 m = eval(f"{k}(x)")
#                 # xr = pure_random(n, shape)
#                 # mr = eval(f"{k}(xr)")
#                 print(f"{k} --> {m}")  # , vs   {mr} for random")
#                 data[k][method] += [m]
#                 if len(data[k]) > 1:
#                     do_plot(k + f"_number{n}_shape{shape}", data[k])
#         # Now let's do a heatmap
#         tab = np.array([[np.average(data[k][method]) for k in list_of_metrics] for method in list_of_methods])
#         print("we have ", tab)
#         heatmap(
#             list_of_methods,
#             list_of_metrics,
#             tab,
#             str(shape) + "_" + str(n) + "_" + str(np.random.randint(50000)) + "_bigartifcompa.png",
#         )


for u in list_of_methods:
    methods[u] = eval(u)


def parallel_create_statistics(n, shape, list_of_methods, list_of_metrics, num=1):
    shape = [int(s) for s in list(shape)]
    for _ in range(num):

        def deal_with_method(method):
            print(f"{method}")
            # x = eval(f"methods[method](n, shape)")
            x = methods[method](n, shape)
            np.array(x).tofile(
                f"pointset_{n}_{shape}_{method}_{default_budget}_{np.random.randint(50000)}.dat".replace(
                    " ", "_"
                )
                .replace("[", " ")
                .replace("]", " ")
            )
            print(f"{method}({n}, {shape}) created in time {default_budget}")
            metrics_values = []
            for k in list_of_metrics:
                # m = eval(f"{k}(x)")
                m = metrics[k](x)
                # print(f"{k} --> {m}")   #, vs   {mr} for random")
                metrics_values += [m]
                print(f"{method}({n}, {shape}) evaluated for {k}")
            return metrics_values

        # for method in list_of_methods:
        results = Parallel(n_jobs=70)(delayed(deal_with_method)(method) for method in list_of_methods)
        for i, method in enumerate(list_of_methods):
            for j, k in enumerate(list_of_metrics):
                data[k][method] += [results[i][j]]

        for k in list_of_metrics:
            if len(data[k]) > 1:
                do_plot(k + f"_number{n}_shape{shape}", data[k])
        # Now let's do a heatmap
        tab = np.array([[np.average(data[k][method]) for k in list_of_metrics] for method in list_of_methods])
        print("we have ", tab)
        # heatmap(list_of_methods, list_of_metrics, tab, "bigartifcompa.png")
        heatmap(
            list_of_methods,
            list_of_metrics,
            tab,
            str(shape) + "_" + str(n) + "_" + str(np.random.randint(50000)) + "_bigartifcompa.png",
        )
        # heatmap(list_of_methods, list_of_metrics, tab, str(shape) + "_" + str(n) + "_bigartifcompa.png")


# size=int(np.random.choice([1,1,1,64,32,16]))
# size=1
# channels = int(np.random.choice([2,3,4]))
# if size == 1:
# channels = np.random.choice([128, 16])
# numpoints = int(np.random.choice([12, 24, 48, 96, 192, 384]))
# print(f"Size={size}, Channels={channels}, numpoints={numpoints}")
# parallel_create_statistics(numpoints, [size,size,channels], num=1, list_of_metrics=list_metrics, list_of_methods=list_of_methods)
##create_statistics(40, [16,16,2])
# quit()
## First we play in dimension two.
# n=30
# for k in list_for_drawing:
#    print(f"Testing {k}")
#    plt.clf()
#    x = eval(f"{k}(n, [1, 2])")
#    assert len(x) == n, f"n={n}, x={x}"
#    for x_ in x:
#        assert 0.99 < np.linalg.norm(x_.flatten()) < 1.01, f"{k} fails with norm {np.linalg.norm(x_.flatten())}"
#    xx = np.array([x[i][0,0] for i in range(len(x))])
#    yy = np.array([x[i][0,1] for i in range(len(x))])
#    indices = np.argsort(xx)
#    xx = xx[indices]
#    yy = yy[indices]
#    nxx = []
#    nyy = []
#    for i in range(len(xx)):
#        nxx += [xx[i], 0]
#        nyy += [yy[i], 0]
#    plt.plot(nxx, nyy, label=k)
#    plt.savefig("sphere_" + k + ".png")
## Then dim 3.
# for k in list_for_drawing:
#    print(f"Testing {k} in dim 3")
#    x = eval(f"{k}(n, [3, 1])")
#    show_points(x, k)


def bigcheck():
    # Now let us go!
    n = 20  # Let us consider batch size 50
    shape = (8, 8, 3)  # Maybe this is what we need for now ?
    for k in [
        "lhs",
        "reduced_jittered",
        "jittered",
        "big_block_symmetry",
        "block_symmetry",
        "greedy_dispersion",
        "dispersion",
        "pure_random",
        "antithetic_pm",
        "dispersion",
        "antithetic_order",
        "antithetic_order_and_sign",
        "dispersion_with_conv",
        "dispersion_with_big_conv",
        "greedy_dispersion_with_big_conv",
        "dispersion_with_mini_conv",
        "greedy_dispersion_with_mini_conv",
        "covering",
        "covering_conv",
        "covering_mini_conv",
    ]:
        print("Starting to play with ", k)
        eval(f"{k}(n, shape)")
        print(f" {k} has been used for generating a batch of {n} points with shape {shape}")


# bigcheck()


def get_a_point_set(n, shape, method=None):
    k = np.random.choice(list_of_methods)
    if method is not None:
        assert method in list_of_methods
        k = method
    print("Working with ", k)
    x = eval(f"{k}({n}, {shape})")
    for i in range(len(x)):
        assert 0.999 < np.linalg.norm(x[i]) < 1.001, "we have norm " + str(np.linalg.norm(x[i]))
    np.array(x).tofile(
        f"pointset_{n}_{shape}_{method}_{default_budget}_{np.random.randint(50000)}.dat".replace(" ", "_")
        .replace("[", " ")
        .replace("]", " ")
    )
    return k, x


# k, x = get_a_point_set(50, (64, 64, 4))


def quasi_randomize(pointset, method):
    n = len(pointset)
    shape = [int(i) for i in list(pointset[0].shape)]
    norms = [np.linalg.norm(pointset[i]) for i in range(n)]
    if method == "none":
        if len(shape) > 1 and shape[0] > 5:
            x = dispersion(n, shape, conv=[int(s / 3) for s in list(shape)[:-1]])
        else:
            x = rs_ng_DiagonalCMA(n, shape)
    else:
        x = get_a_point_set(n, shape, method)
    x = normalize(x)
    for i in range(n):
        x[i] = norms[i] * x[i]
    return x



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np

from . import core


def test_images_adversarial() -> None:
    func = next(core.ImageAdversarial.make_folder_functions(None, model="test"))
    x = np.zeros(func.image.shape)
    value = func(x)
    assert isinstance(value, float)
    assert value < np.inf
    other_func = func.copy().copy()
    value2 = other_func(x)
    assert value2 == value  # same function


def test_image_adversarial_eval() -> None:
    func = next(core.ImageAdversarial.make_folder_functions(None, model="test"))
    output = func.evaluation_function(func.parametrization)
    assert output == 1
    func.targeted = True
    output = func.evaluation_function(func.parametrization)
    assert output == 0


def test_images() -> None:
    func = core.Image()
    x = 7 * np.fabs(np.random.normal(size=func.domain_shape))
    # data = func.parametrization.spawn_child().set_standardized_data(x.flatten()).value
    value = func(x)  # should not touch boundaries, so value should be < np.inf
    assert isinstance(value, float)
    assert value < np.inf
    other_func = func.copy()
    value2 = other_func(x)
    assert value == value2



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import numpy as np
import nevergrad.common.typing as tp
from nevergrad.parametrization import parameter as p


logger = logging.getLogger(__name__)


class CrowdingDistance:
    """This class implements the calculation of crowding distance for NSGA-II."""

    def accumulate_distance_per_objective(self, front: tp.List[p.Parameter], i: int):
        if isinstance(front[0].losses, np.ndarray) and front[0].losses.shape != ():
            is_multiobj: bool = len(front[0].losses) > 1  # isinstance(front[0].loss, np.ndarray)
        else:
            is_multiobj = False
        assert (not is_multiobj and (i == 0)) or is_multiobj

        # Sort the population by objective i
        if is_multiobj:
            front = sorted(front, key=lambda x: x.losses[i])
            objective_minn = front[0].losses[i]
            objective_maxn = front[-1].losses[i]
            assert objective_minn <= objective_maxn

            # Set the crowding distance
            front[0]._meta["crowding_distance"] = float("inf")
            front[-1]._meta["crowding_distance"] = float("inf")

            # All other intermediate solutions are assigned a distance value equal
            # to the absolute normalized difference in the function values of two
            # adjacent solutions.
            for j in range(1, len(front) - 1):
                distance = front[j + 1].losses[i] - front[j - 1].losses[i]

                # Check if minimum and maximum are the same (in which case do nothing)
                if objective_maxn - objective_minn == 0:
                    pass  # undefined
                else:
                    distance = distance / float(objective_maxn - objective_minn)
                logger.debug("front[j]: %s distance: %s", front[j].uid, distance)
                # The overall crowding-distance value is calculated as the sum of
                # individual distance values corresponding to each objective.
                front[j]._meta["crowding_distance"] += distance
        else:
            front = sorted(front, key=lambda x: x.loss)  # type: ignore
            objective_minn = front[0].loss
            objective_maxn = front[-1].loss
            assert objective_minn <= objective_maxn

            # Set the crowding distance
            front[0]._meta["crowding_distance"] = float("inf")
            front[-1]._meta["crowding_distance"] = float("inf")

            # All other intermediate solutions are assigned a distance value equal
            # to the absolute normalized difference in the function values of two
            # adjacent solutions.
            for j in range(1, len(front) - 1):
                if front[j + 1].loss is None or front[j - 1].loss is None:
                    distance = 0
                else:
                    distance = front[j + 1].loss if front[j + 1].loss is not None else 0
                    distance -= front[j - 1].loss if front[j - 1].loss is not None else 0

                # Check if minimum and maximum are the same (in which case do nothing)
                if objective_maxn - objective_minn == 0:
                    pass  # undefined
                else:
                    distance = distance / float(objective_maxn - objective_minn)
                # The overall crowding-distance value is calculated as the sum of
                # individual distance values corresponding to each objective.
                front[j]._meta["crowding_distance"] += distance

    def compute_distance(self, front: tp.List[p.Parameter]):
        """This function assigns the crowding distance to the solutions.
        :param front: The list of solutions.
        """
        size = len(front)

        if size == 0:
            return
        # The boundary solutions (solutions with smallest and largest function values)
        # are set to an infinite (maximum) distance value
        if size == 1:
            front[0]._meta["crowding_distance"] = float("inf")
            return
        if size == 2:
            front[0]._meta["crowding_distance"] = float("inf")
            front[1]._meta["crowding_distance"] = float("inf")
            return

        for f in front:
            f._meta["crowding_distance"] = 0.0

        if isinstance(front[0].losses, np.ndarray) and front[0].losses.shape != ():
            number_of_objectives = len(front[0].losses)
        else:
            number_of_objectives = 1

        for i in range(number_of_objectives):
            self.accumulate_distance_per_objective(front, i)

    def sort(self, candidates: tp.List[p.Parameter], in_place: bool = True) -> tp.List[p.Parameter]:
        if in_place:
            candidates.sort(
                key=lambda elem: elem._meta["crowding_distance"], reverse=True
            )  # Larger -> Less crowded
        return sorted(candidates, key=lambda elem: elem._meta["crowding_distance"], reverse=True)


class FastNonDominatedRanking:
    """Non-dominated ranking of NSGA-II proposed by Deb et al., see [Deb2002]"""

    def compare(self, candidate1: p.Parameter, candidate2: p.Parameter) -> int:
        """Compare the domainance relation of two candidates.

        :param candidate1: Candidate.
        :param candidate2: Candidate.
        """
        one_wins = np.sum(candidate1.losses < candidate2.losses)
        two_wins = np.sum(candidate2.losses < candidate1.losses)
        if one_wins > two_wins:
            return -1
        if two_wins > one_wins:
            return 1
        return 0

    # pylint: disable=too-many-locals
    def compute_ranking(
        self, candidates: tp.List[p.Parameter], k: tp.Optional[int] = None
    ) -> tp.List[tp.List[p.Parameter]]:
        """Compute ranking of candidates.

        :param candidates: List of candidates.
        :param k: Number of individuals.
        """
        n_cand: int = len(candidates)
        # dominated_by_cnt[i]: number of candidates dominating ith candidate
        dominated_by_cnt: tp.List[int] = [0] * n_cand  # [0 for _ in range(len(candidates))]

        # candidates_dominated[i]: List of candidates dominated by ith candidate
        candidates_dominated: tp.List[tp.List[int]] = [[] for _ in range(n_cand)]

        # front[i] contains the list of solutions belonging to front i
        front: tp.List[tp.List[int]] = [[] for _ in range(n_cand + 1)]

        uid2candidate = {c.uid: c for c in candidates}
        uids = [c.uid for c in candidates]
        for c1 in range(n_cand - 1):
            uid1 = uids[c1]
            for c2 in range(c1 + 1, n_cand):
                uid2 = uids[c2]
                dominance_test_result = self.compare(uid2candidate[uid1], uid2candidate[uid2])
                # self.number_of_comparisons += 1
                if dominance_test_result == -1:
                    # c1 wins
                    candidates_dominated[c1].append(c2)
                    dominated_by_cnt[c2] += 1
                elif dominance_test_result == 1:
                    # c2 wins
                    candidates_dominated[c2].append(c1)
                    dominated_by_cnt[c1] += 1

        # Reset rank
        for cand in candidates:
            cand._meta["non_dominated_rank"] = float("inf")

        # Formation of front[0], i.e. candidates that do not dominated by others
        front[0] = [c1 for c1 in range(n_cand) if dominated_by_cnt[c1] == 0]
        last_fronts = 0
        while len(front[last_fronts]) != 0:
            last_fronts += 1
            # Number of candidates in a frontier <= Number of candidates that dominate at least 1 candidate
            assert len(front[last_fronts - 1]) <= len(candidates_dominated)
            for c1 in front[last_fronts - 1]:
                for c2 in candidates_dominated[c1]:
                    dominated_by_cnt[c2] -= 1
                    if dominated_by_cnt[c2] == 0:
                        front[last_fronts].append(c2)

        # Convert index to uid
        # Trim to frontiers that contain the k candidates of interest
        ranked_sublists = []
        count = 0
        for front_i in range(last_fronts):
            count += len(front[front_i])
            for cand_i in front[front_i]:
                uid2candidate[uids[cand_i]]._meta["non_dominated_rank"] = front_i
            ranked_sublists.append([uid2candidate[uids[i]] for i in front[front_i]])
            if (k is not None) and (count >= k):
                break

        return ranked_sublists


def rank(
    population: tp.List[p.Parameter], n_selected: tp.Optional[int] = None
) -> tp.Dict[str, tp.Tuple[int, int, float]]:
    """implements the multi-objective ranking function of NSGA-II."""
    frontier_ranker = FastNonDominatedRanking()
    density_estimator = CrowdingDistance()
    selected_pop: tp.Dict[str, tp.Tuple[int, int, float]] = {}
    frontiers = frontier_ranker.compute_ranking(population)
    count = 0
    next_rank = 0
    for front_i, p_frontier in enumerate(frontiers):
        count += len(p_frontier)
        if n_selected is None or count > n_selected:
            density_estimator.compute_distance(p_frontier)
            density_estimator.sort(p_frontier)
            n_dist_calc = n_selected - len(selected_pop) if n_selected is not None else len(p_frontier)
            for c_i in range(0, n_dist_calc):
                selected_pop[p_frontier[c_i].uid] = (
                    next_rank,
                    front_i,
                    p_frontier[c_i]._meta["crowding_distance"],
                )
                next_rank += 1
            if n_selected is not None:
                break
        if n_selected is not None:
            for candidate in p_frontier:
                selected_pop[candidate.uid] = (next_rank, front_i, float("inf"))
            next_rank += 1
    return selected_pop



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
import numpy as np
from ..common import testing
from . import sequences
from .sequences import samplers


def test_get_first_primes() -> None:
    # check first 10
    first_10 = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29]
    for k in range(11):
        np.testing.assert_array_equal(sequences._get_first_primes(k), first_10[:k])
    # generate a big number of them
    num = np.random.randint(1000, 1000000)
    output = sequences._get_first_primes(num)
    # verify the last value and one random value
    for value in [np.random.choice(output), output[-1]]:
        for k in range(3, 1 + int(np.sqrt(value)), 2):
            assert value % k, f"Value {value} obtained with num={num} can be divided by {k}"


@testing.parametrized(**{name: (sampler,) for name, sampler in samplers.items()})
def test_samplers(sampler_cls: tp.Type[sequences.Sampler]) -> None:
    sampler = sampler_cls(144, 4)
    np.testing.assert_equal(sampler.index, 0)
    output = sampler()
    np.testing.assert_equal(sampler.index, 1)
    np.testing.assert_equal(len(output), 144)
    assert min(output) > 0
    assert max(output) < 1


def test_sampler_draw() -> None:
    sampler = sequences.RandomSampler(5, 4)
    sampler.draw()


@testing.parametrized(
    # lhs=("LHSSampler", [0.069, 0.106, 0.384], [0.282, 0.857, 0.688]),  # previous: for the record
    lhs=("LHSSampler", [0.931, 0.422, 0.391], [0.428, 0.625, 0.797]),
    halton=("HaltonSampler", [0.5, 0.333, 0.2], [0.25, 0.667, 0.4]),
)
def test_sampler_values(name: str, seq1: tp.List[float], seq2: tp.List[float]) -> None:
    budget = 4
    np.random.seed(12)
    sampler = samplers[name](3, budget)
    samples = list(sampler)
    for k, expected in enumerate([seq1, seq2]):
        np.testing.assert_almost_equal(samples[k], expected, decimal=3)
    np.testing.assert_raises(AssertionError, sampler)  # budget is over
    sampler.reinitialize()
    samples2 = list(sampler)
    testing.printed_assert_equal(samples2, samples)


def test_permutation_generator() -> None:
    np.random.seed(12)
    gen = sequences.HaltonPermutationGenerator(5, scrambling=True)
    value1 = list(gen.get_permutations_generator())
    value2 = list(gen.get_permutations_generator())
    gen = sequences.HaltonPermutationGenerator(5, scrambling=True)
    value3 = list(gen.get_permutations_generator())
    testing.printed_assert_equal(value1, value2)
    # testing.printed_assert_equal(value1[:3], [[0, 1], [0, 2, 1], [0, 4, 3, 2, 1]])  # previous: for the record
    testing.printed_assert_equal(value1[:3], [[0, 1], [0, 1, 2], [0, 2, 3, 1, 4]])
    np.testing.assert_raises(AssertionError, testing.printed_assert_equal, value3, value2)
    #
    gen = sequences.HaltonPermutationGenerator(5, scrambling=False)
    value = list(gen.get_permutations_generator())
    testing.printed_assert_equal(value, [np.arange(p) for p in [2, 3, 5, 7, 11]])


def test_rescaler_on_hammersley() -> None:
    np.random.seed(12)
    sampler = sequences.HammersleySampler(dimension=3, budget=4, scrambling=True)
    samples = list(sampler)
    sampler.reinitialize()
    samples2 = list(sampler)
    sampler.reinitialize()
    np.testing.assert_array_equal(
        samples, samples2, "Not repeatable"  # type: ignore
    )  # test repeatability of hammersley first
    rescaler = sequences.Rescaler(sampler)
    sampler.reinitialize()
    rescaled_samples = [rescaler.apply(x) for x in sampler]
    expected = [[1e-15, 0.600, 0.667], [0.333, 0.200, 0.167], [0.667, 1.0, 1e-15], [1.0, 1e-15, 1.0]]
    np.testing.assert_almost_equal(rescaled_samples, expected, decimal=3)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Approximate Rocket Simulation
Based on
https://raw.githubusercontent.com/purdue-orbital/rocket-simulation/master/Simulation2.py
"""


import math
import pyproj
import numpy as np
from nevergrad.parametrization import parameter
from ..base import ArrayExperimentFunction

# pylint: disable=too-many-locals,too-many-statements


class Rocket(ArrayExperimentFunction):
    def __init__(self, symmetry: int = 0) -> None:
        super().__init__(rocket, parametrization=parameter.Array(shape=(24,)), symmetry=symmetry)


def rocket(thrust_bias: np.ndarray):

    assert len(thrust_bias) == 24, "Bad guide length."
    # Covert ang to rads

    def rad(ang):
        return (ang / 360) * 2 * (3.1415926)

    # air density (simple model based on nasa function online)
    def air_density(alt):
        if alt <= 11000:
            T = 15.04 - (0.00649 * alt)
            p = 101.29 * math.pow((T + 273.1) / 288.08, 5.256)
        elif alt <= 25000:
            T = -56.46
            p = float(22.65 * (math.exp(1.73 - 0.000157 * alt))).real
        else:
            T = -131.21 + (0.00299 * alt)
            p = 2.488 * (((T + 273.1) / 216.6) ** -11.388).real
        d = p / (0.2869 * (T + 273.1))
        return d

    def alt(Ex, Ey, Ez):
        ecef = pyproj.Proj(proj="geocent", ellps="WGS84", datum="WGS84")
        lla = pyproj.Proj(proj="latlong", ellps="WGS84", datum="WGS84")
        transformer = pyproj.Transformer.from_proj(ecef, lla)
        _, _, alt = transformer.transform(Ex, Ey, Ez, radians=True)
        return alt

    def grav_force(Ex, Ey, Ez, m):
        # lat = rad(lat)
        G = -6.67408 * (1 / (10**11))  # Gravitational Constant (m^3 kg^-1 s^-2)
        M = 5.972 * (10**24)  # Earth Mass (kg)
        # a = 6398137  # equatoral radius
        # b = 6356752  # polar radius
        # R = math.sqrt((math.pow(math.pow(a, 2) * math.cos(lat), 2) + (math.pow(math.pow(b, 2) * math.sin(lat), 2))) / (
        #            math.pow(a * math.cos(lat), 2) + (math.pow(b * math.sin(lat), 2))))  # Radius of earth (m)
        r = (Ex**2 + Ey**2 + Ez**2) ** 0.5
        F = (G * M * m) / (r**2)  # Force of gravity (N)
        F_z = F * Ez / r
        F_x = F * (Ex / ((Ex**2 + Ey**2) ** 0.5))
        F_y = F * (Ey / ((Ex**2 + Ey**2) ** 0.5))
        return F_x, F_y, F_z  # in the -r direction

    def drag_force(Ex, Ey, Ez, Evx, Evy, Evz):
        cd = 0.94  # coefficent of drag
        a = 0.00487  # cross sectional area m^2
        p = air_density(alt(Ex, Ey, Ez))  # air density with respect to alt
        # drag = (1/2)*p*v_sqrd*cd*a*(vy/(math.sqrt(v)))
        v_sqrd = (Evx**2) + (Evy**2) + (Evz**2)
        if Evx == 0:
            Ex_drag = 0
        else:
            Ex_drag = (1 / 2) * p * v_sqrd * cd * a * (-Evx / (math.sqrt(v_sqrd)))
        if Evy == 0:
            Ey_drag = 0
        else:
            Ey_drag = (1 / 2) * p * v_sqrd * cd * a * (-Evy / (math.sqrt(Evx**2 + Evy**2)))
        if Evz == 0:
            Ez_drag = 0
        else:
            Ez_drag = (1 / 2) * p * v_sqrd * cd * a * (-Evz / (math.sqrt(Evx**2 + Evy**2)))
        return Ex_drag, Ey_drag, Ez_drag

    # Net Force
    def net_force(Ex, Ey, Ez, Evx, Evy, Evz, m):
        Fx_drag, Fy_drag, Fz_drag = drag_force(Ex, Ey, Ez, Evx, Evy, Evz)
        Fx_grav, Fy_grav, Fz_grav = grav_force(Ex, Ey, Ez, m)
        Fx = Fx_drag + Fx_grav
        Fy = Fy_drag + Fy_grav
        Fz = Fz_drag + Fz_grav
        return Fx, Fy, Fz

    # def lift_force        lift -> pitching moment by reference length

    # def side_force        side -> yaw moment by reference length

    # Areodynaminc forces are applied at center of pressfrom math import *

    # Need to find ideal coefficient of drag
    # Diameter assumed 0.3m
    # Burn time 175 s +- 20 s
    # avg thrust 1540 N
    # Total mass 700 kg
    # Propellant mass 1st stage 85% of first stage mass -> 510 kg +- x
    # Total 2nd stage 98 kg
    # 2nd Stage propellant 83 kg

    # GOAL FOR FINAL CODE (this refers to the original code, this does not apply here at least in the foreseeable future)
    # Iterate through launch angels to minimize delta V in the y(rocket frame) direction, to 150km
    # after thrust, maximize delta v in the x(rocket frame) direction
    # Minimize time for initial trust burn as third priority
    # looking for initial launch angle, relative angle after thrust is 0, final velocity

    # Inertial Earth Frame. Reference fram centered at the earth center of gravity (z-up, x-forward, y-right)
    # a = 6398137  # equatoral radius
    # b = 6356752  # polar radius
    # radius = R = math.sqrt((math.pow(math.pow(a, 2) * math.cos(latitude), 2) + (math.pow(math.pow(b, 2) * math.sin(latitude), 2))) / (
    #                 math.pow(a * math.cos(latitude), 2) + (math.pow(b * math.sin(latitude), 2))))  # Radius of earth (m)
    # Ez = radius * sin(latitude)  # Zero at earths center of gravity, going up through north pole
    # Ex = radius * cos(latitude) * cos(longitude)  # Zero at earths center of gravity, going through equator forward
    # Ey = radius * cos(latitude) * sin(longitude)  # Zero at earths center of gravity, going through equator right
    # with open('LLA.csv', 'r') as f:
    # #     next(f)
    # #     reader = csv.reader(f, 'excel')
    # #     data = []
    # #     for row in reader:
    # #         data.append(float(row[0]))
    # #     latitude = rad(data[0])
    # #     longitude = rad(data[1])
    # #     altitude = data[2]

    # This is not the same as in the original code (just minor modifications).
    altitude = float(0)
    latitude = rad(float(28.5729))
    longitude = rad(float(80.659))
    # Altitude,Latitude,Longitude
    # 0,28.5729,80.659

    # Black Rock Navada. From 25km. 45 degrees from out of earth, due east.
    # latitude = rad(28.5729)  # N. Latitude
    # longitude = rad(80.6490)  # W. Longitude
    # altitude = 0    # Altitude of rocket in meters
    ecef = pyproj.Proj(proj="geocent", ellps="WGS84", datum="WGS84")
    lla = pyproj.Proj(proj="latlong", ellps="WGS84", datum="WGS84")
    transformer = pyproj.Transformer.from_proj(lla, ecef)
    Ex, Ey, Ez = transformer.transform(longitude, latitude, altitude, radians=True)
    Evx, Evy, Evz = 0, 0, 0
    r_initial = (Ex**2 + Ey**2 + Ez**2) ** 0.5
    # print(Ex, Ey, Ez, r_initial, sep="\t")

    # Rocket specs
    roc_mass = 0.0472  # mass of rocket in kg (not including engine
    theta = 45  # Angle of launch from z
    phi = 45  # Angle of launch from x
    # Original code: eng_file = "C6.csv"  # engine file location
    eng_mass_initial = 0.024  # Loaded engine mass in kg
    eng_mass_final = 0.0132  # empty engine mass in kg
    total_mass = roc_mass + eng_mass_initial
    final_roc_mass = roc_mass + eng_mass_final

    # Earth rotation speed at inital position
    # Position Velocity Attitude(Earth Centric x-forward y-right z-up)
    # latitude and longitude position

    # Adapted from the orignal code (minor modifications).
    thrust = np.asarray(
        [
            [0.0, 0.0],
            [0.946, 0.031],
            [4.826, 0.092],
            [9.936, 0.139],
            [14.09, 0.192],
            [11.446, 0.209],
            [7.381, 0.231],
            [6.151, 0.248],
            [5.489, 0.292],
            [4.921, 0.37],
            [4.448, 0.475],
            [4.258, 0.671],
            [4.542, 0.702],
            [4.164, 0.723],
            [4.448, 0.85],
            [4.353, 1.063],
            [4.353, 1.211],
            [4.069, 1.242],
            [4.258, 1.303],
            [4.353, 1.468],
            [4.448, 1.656],
            [4.448, 1.821],
            [2.933, 1.834],
            [1.325, 1.847],
            [0.0, 1.86],
        ]
    )

    thrust_list = np.asarray([thrust[int(i)][0] for i in range(len(thrust) - 1)])
    thrust_time_list = np.asarray([thrust[i + 1][1] - thrust[i][1] for i in range(0, len(thrust) - 1)])
    total_thrust = np.sum(np.multiply(thrust_list, thrust_time_list))

    # We moodify the thrust while preserving the sum (this is an adaptation to Nevergrad).
    # 1: we modify.
    thrust_list = np.multiply(thrust_list, np.exp(thrust_bias))
    # 2: we normalize.
    thrust_list = thrust_list * total_thrust / np.sum(np.multiply(thrust_list, thrust_time_list))

    for i in range(len(thrust) - 1):
        thrust[i][0] = thrust_list[i]
    # total_mass vs time curve
    # this is used to represent the mass loss while the rocket burns fuel
    mass_time = []
    # total_thrust = 0
    # for row in thrust:
    #    total_thrust += row[0]

    mass_loss = eng_mass_initial - eng_mass_final
    mass_reman = eng_mass_initial
    for row in thrust:
        # Equation below weird to me: this is not normalized by time ? the mass which is lost should be proportional
        # to thrust x delta-time, right ?
        # percentage = row[0] / total_thrust   # percentage of total thrust to find percentage of mass lost
        percentage = row[0] / np.sum(
            thrust_list
        )  # percentage of total thrust to find percentage of mass lost
        assert percentage >= -1e-5, f"percentage is {percentage}"
        assert percentage <= 1.0
        mass_loss = mass_reman * percentage
        mass_reman -= mass_loss
        total_mass = roc_mass + mass_reman
        mass_time.append([total_mass, row[1]])
    mass_list = [mass_time[i][0] for i in range(0, len(thrust))]

    # Lists used to store data and later import data to excel
    Ex_list, Ey_list, Ez_list = np.asarray([Ex]), np.asarray([Ey]), np.asarray([Ez])
    Evx_list, Evy_list, Evz_list = np.asarray([Evx]), np.asarray([Evy]), np.asarray([Evz])
    time_list = np.asarray([0])
    r_list = np.asarray([(Ex**2 + Ey**2 + Ez**2) ** 0.5])

    # Initializing variables
    time = 0.0  # time in seconds

    # while thrust is greater than zero
    # this is while the rocket engine is firing
    for i in range(len(thrust) - 2):
        r = (Ex**2 + Ey**2 + Ez**2) ** 0.5
        dt = thrust[i][1]
        Efx, Efy, Efz = net_force(Ex, Ey, Ez, Evx, Evy, Evz, mass_list[i])
        Ex += Evx * dt
        Ey += Evy * dt
        Ez += Evz * dt
        dt = thrust[i + 1][1] - thrust[i][1]
        Evz += (((thrust[i][0] * math.cos(theta)) + Efz) * dt) / mass_list[i]
        Evx += (((thrust[i][0] * math.sin(theta) * math.cos(phi)) + Efx) * dt) / mass_list[i]
        Evy += (((thrust[i][0] * math.sin(theta) * math.sin(phi)) + Efy) * dt) / mass_list[i]
        time += dt
        Ex_list = np.append(Ex_list, (round(Ex, 6)))
        Ey_list = np.append(Ey_list, (round(Ey, 6)))
        Ez_list = np.append(Ez_list, (round(Ez, 6)))
        Evx_list = np.append(Evx_list, (round(Evx, 6)))
        Evy_list = np.append(Evy_list, (round(Evy, 6)))
        Evz_list = np.append(Evz_list, (round(Evz, 6)))
        time_list = np.append(time_list, (round(time, 6)))
        r_list = np.append(r_list, (round(r, 6)))

    # After thrust
    # This is when the engine is out of fuel and there is no longer a thrust force
    time_step = 0.05  # time time_step in seconds
    dt = time_step
    while r > r_initial:
        r = (Ex**2 + Ey**2 + Ez**2) ** 0.5
        Efx, Efy, Efz = net_force(Ex, Ey, Ez, Evx, Evy, Evz, final_roc_mass)
        Ex += Evx * dt
        Ey += Evy * dt
        Ez += Evz * dt
        Evx += (Efx * dt) / final_roc_mass
        Evy += (Efy * dt) / final_roc_mass
        Evz += (Efz * dt) / final_roc_mass
        # print(Evx, Evy, Evz, r, sep='\t')
        time += dt
        # update(Ex, Ey, Ez, Evx, Evy, Evz, time, r)
        Ex_list = np.append(Ex_list, (round(Ex, 6)))
        Ey_list = np.append(Ey_list, (round(Ey, 6)))
        Ez_list = np.append(Ez_list, (round(Ez, 6)))
        Evx_list = np.append(Evx_list, (round(Evx, 6)))
        Evy_list = np.append(Evy_list, (round(Evy, 6)))
        Evz_list = np.append(Evz_list, (round(Evz, 6)))
        time_list = np.append(time_list, (round(time, 6)))
        r_list = np.append(r_list, (round(r, 6)))

    return 1.0 - max(Ez_list) / 3032708.353202  # Should be 0 for input (0.,....,0.)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path
import PIL.Image
import numpy as np
import pytest
from nevergrad.functions.images import imagelosses


def test_reference() -> None:
    assert imagelosses.SumAbsoluteDifferences.REQUIRES_REFERENCE
    assert not imagelosses.Blur.REQUIRES_REFERENCE
    assert not imagelosses.Koncept512.REQUIRES_REFERENCE
    assert not imagelosses.Brisque.REQUIRES_REFERENCE
    assert len([loss for loss in imagelosses.registry.values() if loss.REQUIRES_REFERENCE]) == 5
    assert len([loss for loss in imagelosses.registry.values() if not loss.REQUIRES_REFERENCE]) == 3


def test_l1_loss() -> None:
    loss = imagelosses.SumAbsoluteDifferences(reference=124.0 * np.ones((300, 400, 3)))
    assert loss(np.ones((300, 400, 3))) == 44280000.0


@pytest.mark.parametrize("loss_name", imagelosses.registry)  # type: ignore
def test_consistency_losses_with_oteytaud(loss_name: str) -> None:
    path = Path(__file__).with_name("headrgb_olivier.png")
    image = PIL.Image.open(path).resize((256, 256), PIL.Image.BICUBIC)
    data = np.asarray(image)[:, :, :3]  # 4th Channel is pointless here, only 255.

    data_flip = np.flip(data, 0).copy()  # Copy necessary as some nets do not support negative stride.
    loss_class = imagelosses.registry[loss_name]
    loss = loss_class(reference=data)
    random_data = np.random.uniform(low=0.0, high=255.0, size=data.shape)
    loss_data = loss(data)
    assert loss_data < 1000.0
    assert loss_data > -1000.0
    assert loss_data == loss(data)
    assert isinstance(loss_data, float)
    loss_random_data = loss(random_data)
    assert loss_random_data == loss(random_data)
    assert "Blur" in loss_name or loss_data < loss_random_data, f"Loss {loss_name} fails on oteytaud's photo."
    if loss.REQUIRES_REFERENCE:
        assert loss_data == 0.0
        if "Histogram" not in loss_name:
            assert loss_data < loss(data_flip)
        else:
            assert loss_data == loss(data_flip)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import time
import warnings
import inspect
import datetime
import logging
from pathlib import Path
import numpy as np
import nevergrad.common.typing as tp
from nevergrad.common import errors
from nevergrad.parametrization import parameter as p
from nevergrad.parametrization import helpers
from . import base

global_logger = logging.getLogger(__name__)


class OptimizationPrinter:
    """Printer to register as callback in an optimizer, for printing
    best point regularly.

    Parameters
    ----------
    print_interval_tells: int
        max number of evaluation before performing another print
    print_interval_seconds: float
        max number of seconds before performing another print
    """

    def __init__(self, print_interval_tells: int = 1, print_interval_seconds: float = 60.0) -> None:
        assert print_interval_tells > 0
        assert print_interval_seconds > 0
        self._print_interval_tells = int(print_interval_tells)
        self._print_interval_seconds = print_interval_seconds
        self._next_tell = self._print_interval_tells
        self._next_time = time.time() + print_interval_seconds

    def __call__(self, optimizer: base.Optimizer, *args: tp.Any, **kwargs: tp.Any) -> None:
        if time.time() >= self._next_time or self._next_tell >= optimizer.num_tell:
            self._next_time = time.time() + self._print_interval_seconds
            self._next_tell = optimizer.num_tell + self._print_interval_tells
            x = optimizer.provide_recommendation()
            print(f"After {optimizer.num_tell}, recommendation is {x}")  # TODO fetch value


class OptimizationLogger:
    """Logger to register as callback in an optimizer, for Logging
    best point regularly.

    Parameters
    ----------
    logger:
        given logger that callback will use to log
    log_level:
        log level that logger will write to
    log_interval_tells: int
        max number of evaluation before performing another log
    log_interval_seconds:
        max number of seconds before performing another log
    """

    def __init__(
        self,
        *,
        logger: logging.Logger = global_logger,
        log_level: int = logging.INFO,
        log_interval_tells: int = 1,
        log_interval_seconds: float = 60.0,
    ) -> None:
        assert log_interval_tells > 0
        assert log_interval_seconds > 0
        self._logger = logger
        self._log_level = log_level
        self._log_interval_tells = int(log_interval_tells)
        self._log_interval_seconds = log_interval_seconds
        self._next_tell = self._log_interval_tells
        self._next_time = time.time() + log_interval_seconds

    def __call__(self, optimizer: base.Optimizer, *args: tp.Any, **kwargs: tp.Any) -> None:
        if time.time() >= self._next_time or self._next_tell >= optimizer.num_tell:
            self._next_time = time.time() + self._log_interval_seconds
            self._next_tell = optimizer.num_tell + self._log_interval_tells
            if optimizer.num_objectives == 1:
                x = optimizer.provide_recommendation()
                self._logger.log(self._log_level, "After %s, recommendation is %s", optimizer.num_tell, x)
            else:
                losses = optimizer._hypervolume_pareto.get_min_losses()  # type: ignore
                self._logger.log(
                    self._log_level,
                    "After %s, the respective minimum loss for each objective in the pareto front is %s",
                    optimizer.num_tell,
                    losses,
                )


class ParametersLogger:
    """Logs parameter and run information throughout into a file during
    optimization.

    Parameters
    ----------
    filepath: str or pathlib.Path
        the path to dump data to
    append: bool
        whether to append the file (otherwise it replaces it)
    order: int
        order of the internal/model parameters to extract

    Example
    -------

    .. code-block:: python

        logger = ParametersLogger(filepath)
        optimizer.register_callback("tell",  logger)
        optimizer.minimize()
        list_of_dict_of_data = logger.load()

    Note
    ----
    Arrays are converted to lists
    """

    def __init__(self, filepath: tp.Union[str, Path], append: bool = True, order: int = 1) -> None:
        self._session = datetime.datetime.now().strftime("%y-%m-%d %H:%M:%S")
        self._filepath = Path(filepath)
        self._order = order
        if self._filepath.exists() and not append:
            self._filepath.unlink()  # missing_ok argument added in python 3.8
        self._filepath.parent.mkdir(exist_ok=True, parents=True)

    def __call__(self, optimizer: base.Optimizer, candidate: p.Parameter, loss: tp.FloatLoss) -> None:
        data = {
            "#parametrization": optimizer.parametrization.name,
            "#optimizer": optimizer.name,
            "#session": self._session,
            "#num-ask": optimizer.num_ask,
            "#num-tell": optimizer.num_tell,
            "#num-tell-not-asked": optimizer.num_tell_not_asked,
            "#uid": candidate.uid,
            "#lineage": candidate.heritage["lineage"],
            "#generation": candidate.generation,
            "#parents_uids": [],
            "#loss": loss,
        }
        if optimizer.num_objectives > 1:  # multiobjective losses
            data.update({f"#losses#{k}": val for k, val in enumerate(candidate.losses)})
            data["#pareto-length"] = len(optimizer.pareto_front())
        if hasattr(optimizer, "_configured_optimizer"):
            configopt = optimizer._configured_optimizer  # type: ignore
            if isinstance(configopt, base.ConfiguredOptimizer):
                data.update({"#optimizer#" + x: str(y) for x, y in configopt.config().items()})
        if isinstance(candidate._meta.get("sigma"), float):
            data["#meta-sigma"] = candidate._meta["sigma"]  # for TBPSA-like algorithms
        if candidate.generation > 1:
            data["#parents_uids"] = candidate.parents_uids
        for name, param in helpers.flatten(candidate, with_containers=False, order=1):
            val = param.value
            if isinstance(val, (np.float_, np.int_, np.bool_)):
                val = val.item()
            if inspect.ismethod(val):
                val = repr(val.__self__)  # show mutation class
            data[name if name else "0"] = val.tolist() if isinstance(val, np.ndarray) else val
            if isinstance(param, p.Data):
                val = param.sigma.value
                data[(name if name else "0") + "#sigma"] = (
                    val.tolist() if isinstance(val, np.ndarray) else val
                )
        try:  # avoid bugging as much as possible
            with self._filepath.open("a") as f:
                f.write(json.dumps(data) + "\n")
        except Exception as e:  # pylint: disable=broad-except
            warnings.warn(f"Failing to json data: {e}")

    def load(self) -> tp.List[tp.Dict[str, tp.Any]]:
        """Loads data from the log file"""
        data: tp.List[tp.Dict[str, tp.Any]] = []
        if self._filepath.exists():
            with self._filepath.open("r") as f:
                for line in f.readlines():
                    data.append(json.loads(line))
        return data

    def load_flattened(self, max_list_elements: int = 24) -> tp.List[tp.Dict[str, tp.Any]]:
        """Loads data from the log file, and splits lists (arrays) into multiple arguments

        Parameters
        ----------
        max_list_elements: int
            Maximum number of elements displayed from the array, each element is given a
            unique id of type list_name#i0_i1_...
        """
        data = self.load()
        flat_data: tp.List[tp.Dict[str, tp.Any]] = []
        for element in data:
            list_keys = {key for key, val in element.items() if isinstance(val, list)}
            flat_data.append({key: val for key, val in element.items() if key not in list_keys})
            for key in list_keys:
                for k, (indices, value) in enumerate(np.ndenumerate(element[key])):
                    if k >= max_list_elements:
                        break
                    flat_data[-1][key + "#" + "_".join(str(i) for i in indices)] = value
        return flat_data

    def to_hiplot_experiment(
        self, max_list_elements: int = 24
    ) -> tp.Any:  # no typing here since Hiplot is not a hard requirement
        """Converts the logs into an hiplot experiment for display.

        Parameters
        ----------
        max_list_elements: int
            maximum number of elements of list/arrays to export (only the first elements are extracted)

        Example
        -------
        .. code-block:: python

            exp = logs.to_hiplot_experiment()
            exp.display(force_full_width=True)

        Note
        ----
        - You can easily change the axes of the XY plot:
          :code:`exp.display_data(hip.Displays.XY).update({'axis_x': '0#0', 'axis_y': '0#1'})`
        - For more context about hiplot, check:

          - blogpost: https://ai.facebook.com/blog/hiplot-high-dimensional-interactive-plots-made-easy/
          - github repo: https://github.com/facebookresearch/hiplot
          - documentation: https://facebookresearch.github.io/hiplot/
        """
        # pylint: disable=import-outside-toplevel
        try:
            import hiplot as hip
        except ImportError as e:
            raise ImportError(
                f"{self.__class__.__name__} requires hiplot which is not installed by default "
                "(pip install hiplot)"
            ) from e
        exp = hip.Experiment()
        for xp in self.load_flattened(max_list_elements=max_list_elements):
            dp = hip.Datapoint(
                from_uid=xp.get("#parents_uids#0"),
                uid=xp["#uid"],
                values={
                    x: y for x, y in xp.items() if not (x.startswith("#") and ("uid" in x or "ask" in x))
                },
            )
            exp.datapoints.append(dp)
        exp.display_data(hip.Displays.XY).update({"axis_x": "#num-tell", "axis_y": "#loss"})
        # for the record, some more options:
        exp.display_data(hip.Displays.XY).update({"lines_thickness": 1.0, "lines_opacity": 1.0})
        return exp


class OptimizerDump:
    """Dumps the optimizer to a pickle file at every call.

    Parameters
    ----------
    filepath: str or Path
        path to the pickle file
    """

    def __init__(self, filepath: tp.Union[str, Path]) -> None:
        self._filepath = filepath

    def __call__(self, opt: base.Optimizer, *args: tp.Any, **kwargs: tp.Any) -> None:
        opt.dump(self._filepath)


class ProgressBar:
    """Progress bar to register as callback in an optimizer"""

    def __init__(self) -> None:
        self._progress_bar: tp.Any = None
        self._current = 0

    def __call__(self, optimizer: base.Optimizer, *args: tp.Any, **kwargs: tp.Any) -> None:
        if self._progress_bar is None:
            # pylint: disable=import-outside-toplevel
            try:
                from tqdm import tqdm  # Inline import to avoid additional dependency
            except ImportError as e:
                raise ImportError(
                    f"{self.__class__.__name__} requires tqdm which is not installed by default "
                    "(pip install tqdm)"
                ) from e
            self._progress_bar = tqdm()
            self._progress_bar.total = optimizer.budget
            self._progress_bar.update(self._current)
        self._progress_bar.update(1)
        self._current += 1

    def __getstate__(self) -> tp.Dict[str, tp.Any]:
        """Used for pickling (tqdm is not picklable)"""
        state = dict(self.__dict__)
        state["_progress_bar"] = None
        return state


class EarlyStopping:
    """Callback for stopping the :code:`minimize` method before the budget is
    fully used.

    Parameters
    ----------
    stopping_criterion: func(optimizer) -> bool
        function that takes the current optimizer as input and returns True
        if the minimization must be stopped

    Note
    ----
    This callback must be register on the "ask" method only.

    Example
    -------
    In the following code, the :code:`minimize` method will be stopped at the 4th "ask"

    >>> early_stopping = ng.callbacks.EarlyStopping(lambda opt: opt.num_ask > 3)
    >>> optimizer.register_callback("ask", early_stopping)
    >>> optimizer.minimize(_func, verbosity=2)

    A couple other options (equivalent in case of non-noisy optimization) for stopping
    if the loss is below 12:

    >>> early_stopping = ng.callbacks.EarlyStopping(lambda opt: opt.recommend().loss < 12)
    >>> early_stopping = ng.callbacks.EarlyStopping(lambda opt: opt.current_bests["minimum"].mean < 12)
    """

    def __init__(self, stopping_criterion: tp.Callable[[base.Optimizer], bool]) -> None:
        self.stopping_criterion = stopping_criterion

    def __call__(self, optimizer: base.Optimizer, *args: tp.Any, **kwargs: tp.Any) -> None:
        if args or kwargs:
            raise errors.NevergradRuntimeError("EarlyStopping must be registered on ask method")
        if self.stopping_criterion(optimizer):
            raise errors.NevergradEarlyStopping("Early stopping criterion is reached")

    @classmethod
    def timer(cls, max_duration: float) -> "EarlyStopping":
        """Early stop when max_duration seconds has been reached (from the first ask)"""
        return cls(_DurationCriterion(max_duration))

    @classmethod
    def no_improvement_stopper(cls, tolerance_window: int) -> "EarlyStopping":
        """Early stop when loss didn't reduce during tolerance_window asks"""
        return cls(_LossImprovementToleranceCriterion(tolerance_window))


class _DurationCriterion:
    def __init__(self, max_duration: float) -> None:
        self._start = float("inf")
        self._max_duration = max_duration

    def __call__(self, optimizer: base.Optimizer) -> bool:
        if np.isinf(self._start):
            self._start = time.time()
        return time.time() > self._start + self._max_duration


class _LossImprovementToleranceCriterion:
    def __init__(self, tolerance_window: int) -> None:
        self._tolerance_window: int = tolerance_window
        self._best_value: tp.Optional[np.ndarray] = None
        self._tolerance_count: int = 0

    def __call__(self, optimizer: base.Optimizer) -> bool:
        best_param = optimizer.provide_recommendation()
        if best_param is None or (best_param.loss is None and best_param._losses is None):
            return False
        best_last_losses = best_param.losses
        if self._best_value is None:
            self._best_value = best_last_losses
            return False
        if self._best_value <= best_last_losses:
            self._tolerance_count += 1
        else:
            self._tolerance_count = 0
            self._best_value = best_last_losses
        return self._tolerance_count > self._tolerance_window
